from itertools import chain

import meshio

from ada.core.containers import Nodes
from ada.core.utils import Counter
from ada.fem import FEM, Elem
from ada.fem.containers import FemElements


def meshio_read_fem(assembly, fem_file, fem_name=None):
    """
    Import a FEM file using the meshio package.


    :param assembly: Assembly object
    :param fem_file: Path to fem file
    :param fem_name: Name of FEM model
    :type assembly: ada.Assembly
    """
    from ada import Node, Part

    from . import meshio_to_ada_type

    mesh = meshio.read(fem_file)
    name = fem_name if fem_name is not None else "Part-1"
    fem = FEM(name)

    def to_node(data):
        return Node(data[1], data[0])

    point_ids = mesh.points_id if "points_id" in mesh.__dict__.keys() else [i + 1 for i, x in enumerate(mesh.points)]
    elem_counter = Counter(1)

    cell_ids = (
        mesh.cells_id
        if "cells_id" in mesh.__dict__.keys()
        else [[next(elem_counter) for cell in cellblock.data] for cellblock in mesh.cells]
    )
    fem._nodes = Nodes([to_node(p) for p in zip(point_ids, mesh.points)])

    cell_block_counter = Counter(0)

    def to_elem(cellblock):
        block_id = next(cell_block_counter)
        return [
            Elem(
                cell_ids[block_id][i],
                [fem.nodes.from_id(point_ids[c]) for c in cell],
                meshio_to_ada_type[cellblock.type],
            )
            for i, cell in enumerate(cellblock.data)
        ]

    fem._elements = FemElements(chain.from_iterable(map(to_elem, mesh.cells)))
    assembly.add_part(Part(name, fem=fem))



import numpy as np
from pythreejs import (
    BufferAttribute,
    BufferGeometry,
    LineBasicMaterial,
    LineSegments,
    Mesh,
    MeshBasicMaterial,
    Points,
    PointsMaterial,
)

from .common import format_color


def edges_to_mesh(name, np_edge_vertices, np_edge_indices, edge_color, linewidth=1):
    """

    :param name:
    :param np_edge_vertices:
    :param np_edge_indices:
    :param edge_color:
    :param linewidth:
    :return:
    :rtype: pythreejs.objects.LineSegments_autogen.LineSegments
    """
    edge_geometry = BufferGeometry(
        attributes={
            "position": BufferAttribute(np_edge_vertices),
            "index": BufferAttribute(np_edge_indices),
        }
    )
    edge_material = LineBasicMaterial(color=format_color(*edge_color), linewidth=linewidth)

    edge_geom = LineSegments(
        geometry=edge_geometry,
        material=edge_material,
        type="LinePieces",
        name=name,
    )

    return edge_geom


def vertices_to_mesh(name, vertices, vertex_color, vertex_width=5):
    """

    :param name:
    :param vertices:
    :param vertex_color: RGB tuple (r, g, b)
    :param vertex_width:
    :return:
    :rtype: pythreejs.objects.Points_autogen.Points
    """
    vertices = np.array(vertices, dtype=np.float32)
    attributes = {"position": BufferAttribute(vertices, normalized=False)}
    mat = PointsMaterial(color=format_color(*vertex_color), sizeAttenuation=False, size=vertex_width)
    geom = BufferGeometry(attributes=attributes)
    points = Points(geometry=geom, material=mat, name=name)
    return points


def faces_to_mesh(name, vertices, faces, colors, opacity=None):
    """

    :param name:
    :param vertices:
    :param faces:
    :param colors:
    :param opacity:
    :return:
    """
    geometry = BufferGeometry(
        attributes=dict(
            position=BufferAttribute(vertices, normalized=False),
            index=BufferAttribute(faces, normalized=False),
            color=BufferAttribute(colors),
        )
    )

    mat_atts = dict(vertexColors="VertexColors", side="DoubleSide")
    if opacity is not None:
        mat_atts["opacity"] = opacity
        mat_atts["transparent"] = True

    material = MeshBasicMaterial(**mat_atts)
    mesh = Mesh(
        name=name,
        geometry=geometry,
        material=material,
    )
    return mesh



import numpy as np


class MatDataSheet:
    """

    :param name:
    :param mat_class:
    :param mat_type:
    :param testdata:
    """

    def __init__(self, name, mat_class, mat_type, testdata):
        self._name = name
        self._class = mat_class
        self._type = mat_type
        self._test_data = testdata

    @property
    def name(self):
        return self._name

    @property
    def mat_class(self):
        return self._class

    @property
    def type(self):
        return self._type

    @property
    def test_data(self):
        return self._test_data

    @property
    def uniaxial(self):
        for td in self.test_data:
            if "uniaxial" in td.load_type.lower():
                return td
        return None

    @property
    def biaxial(self):
        for td in self.test_data:
            if "biaxial" in td.load_type.lower():
                return td
        return None

    @property
    def planar(self):
        for td in self.test_data:
            if "planar" in td.load_type.lower():
                return td
        return None

    @property
    def visualize_eng(self):
        from ada.core.utils import easy_plotly

        in_data = {}
        for d in self.test_data:
            in_data[f"{d.name} ({d.load_type})"] = (d.e, d.S)
        return easy_plotly(
            self.name,
            in_data,
            mode="lines+markers",
            xlbl="Engineering Strain",
            ylbl="Engineering Stress [MPa]",
        )

    @property
    def visualize_true(self):
        in_data = dict()
        from ada.core.utils import easy_plotly

        for test in self.test_data:
            in_data[test.title] = (test.eps, test.sig)
        return easy_plotly(
            self.name,
            in_data,
            mode="lines+markers",
            xlbl="True Strain",
            ylbl="True Stress [MPa]",
        )

    @property
    def true_dict(self):
        in_data = dict()
        for d in self.test_data:
            in_data[f"{d.name} ({d.load_type})"] = (d.eps, d.sig)
        return in_data

    @property
    def eng_dict(self):
        in_data = dict()
        for d in self.test_data:
            in_data[f"{d.name} ({d.load_type})"] = (d.e, d.S)
        return in_data

    def __repr__(self):
        tests = "\n".join([str(x) for x in self.test_data])
        return f'{self.name}\nClass: "{self.mat_class}", Type: "{self.type}"\nTest Data:\n{tests}'


class TestData:
    def __init__(self, mat_data_path, incompressible=True):
        self._name = None
        self._standard = None
        self._temperature = None
        self._load_type = None
        self._source = None
        self._created_by = None
        self._incompressible = incompressible
        with open(mat_data_path, "r") as d:
            lines = d.readlines()
        comments = ""
        data = []
        cols = []
        for li in lines:
            if "#" in li[0]:
                if "created by" in li.lower():
                    self._created_by = li.split(":")[-1].strip()
                if "column" in li:
                    cols.append(li.split(":")[-1].strip())
                if "standard" in li.lower():
                    self._standard = li.split(":")[-1].strip() if "none" not in li.lower() else None
                if "temperature" in li.lower():
                    self._temperature = li.split(":")[-1].strip()
                if "load type" in li.lower():
                    self._load_type = li.split(":")[-1].strip()
                if "source" in li.lower():
                    self._source = li.split(":")[-1].strip()
                if "name" in li.lower():
                    self._name = li.split(":")[-1].strip()
                comments += "".join(li[1:])

            else:
                d = li.split(",")
                data.append([float(x) for x in d])

        self._data = list(zip(*data))
        self._cols = cols
        self._comments = comments

    @property
    def standard(self):
        return self._standard

    @property
    def load_type(self):
        return self._load_type

    @property
    def comments(self):
        return self._comments

    @property
    def cols(self):
        return self._cols

    @property
    def data(self):
        return self._data

    @property
    def eps(self):
        """

        :return: True strain
        """
        for i, c in enumerate(self.cols):
            if "true strain" in c.lower():
                return np.array([self.data[i]])

        if self.e is not None:
            return np.array([np.log(1 + e) for e in self.e])

        return None

    @property
    def sig(self):
        """

        :return: True stress
        """
        for i, c in enumerate(self.cols):
            if "true stress" in c.lower():
                return np.array(self.data[i])

        if self.S is not None:
            return np.array([s * (1 + eps) for eps, s in zip(self.eps, self.S)])

        return None

    @property
    def e(self):
        """

        :return: Engineering strain
        """
        for i, c in enumerate(self.cols):
            if "engineering strain" in c.lower():
                return np.array(self.data[i])

        if self.eps is not None:
            if self._incompressible is True:
                return np.exp(self.eps) - 1

        return None

    @property
    def S(self):
        """

        :return: Engineering stress
        """
        for i, c in enumerate(self.cols):
            if "engineering stress" in c.lower():
                return np.array(self.data[i])

        if self.sig is not None:
            return np.array([sig / (1 + eps) for eps, sig in (self.eps, self.sig)])
        return None

    @property
    def name(self):
        return self._name

    @property
    def title(self):
        if self.standard is not None:
            return self.standard + f" ({self.load_type}) Engineering and True stress-strain curves"
        else:
            return self._name + f" ({self.load_type}) Engineering and True stress-strain curves"

    @property
    def visualize(self):
        from ada.core.utils import easy_plotly

        in_data = {"Engineering": (self.e, self.S), "True": (self.eps, self.sig)}
        return easy_plotly(self.title, in_data, xlbl="Strain [-]", ylbl="Stress [Pa]")

    def __repr__(self):
        cols = ", ".join([f'"{e}"' for e in self.cols])
        return f"{self.standard}: {self.load_type} (columns: {cols})"



import logging

import numpy as np

from . import constants as co
from .containers import FemElements, FemSections, FemSets

__all__ = [
    "Amplitude",
    "Bc",
    "Csys",
    "InteractionProperty",
    "Interaction",
    "ElemShapes",
    "Step",
    "Surface",
    "Elem",
    "Connector",
    "Constraint",
    "PredefinedField",
    "FemSet",
    "FEM",
    "Mass",
    "HistOutput",
    "FieldOutput",
    "ConnectorSection",
    "Load",
    "LoadCase",
    "FemSection",
    "Spring",
]


class FemBase:
    """

    :param name:
    :param metadata:
    :param parent:
    """

    def __init__(self, name, metadata, parent):
        self.name = name
        self.parent = parent
        self._metadata = metadata if metadata is not None else dict()

    @property
    def name(self):
        """

        :return:
        """
        return self._name

    @name.setter
    def name(self, value):
        if str.isnumeric(value[0]):
            raise ValueError("Name cannot start with numeric")
        self._name = value

    @property
    def parent(self):
        """

        :rtype: ada.fem.FEM
        """
        return self._parent

    @parent.setter
    def parent(self, value):
        # if type(value) not in (FEM, Step):
        #     raise ValueError(f'Parent type "{type(value)}" is not supported')
        self._parent = value

    @property
    def metadata(self):
        """

        :return:
        """
        return self._metadata

    @property
    def on_assembly_level(self):
        """

        :return:
        """
        # TODO: This is not really working correctly. This must be fixed
        from ada import Assembly

        return True if type(self.parent.parent) is Assembly else False

    @property
    def instance_name(self):
        """

        :return:
        """
        if self.on_assembly_level is False:
            return self.name
        else:
            return self.parent.instance_name + "." + self.name


class ElemShapes:
    """

    :param el_type:
    """

    # 2D elements
    tri = ["S3", "S3R", "R3D3"]
    quad = ["S4", "S4R", "R3D4"]
    quad8 = ["S8", "S8R"]
    quad6 = ["STRI65"]
    shell = tri + quad + quad8 + quad6
    # 3D elements
    cube8 = ["C3D8", "C3D8R", "C3D8H"]
    cube20 = ["C3D20", "C3D20R", "C3D20RH"]
    cube27 = ["C3D27"]
    tetrahedron = ["C3D4"]
    tetrahedron10 = ["C3D10"]
    pyramid5 = ["C3D5", "C3D5H"]
    prism6 = ["C3D6"]
    prism15 = ["C3D15"]
    volume = cube8 + cube20 + tetrahedron10 + tetrahedron + pyramid5 + prism15 + prism6
    # 1D/0D elements
    bm2 = ["B31", "B32"]
    beam = bm2
    spring1n = ["SPRING1"]
    spring2n = ["SPRING2"]
    springs = spring1n + spring2n
    masses = ["MASS", "ROTARYI"]
    connectors = ["CONNECTOR", "CONN3D2"]
    other2n = connectors
    other = other2n

    @staticmethod
    def is_valid_elem(elem_type):
        value = elem_type.upper()
        if (
            value
            not in ElemShapes.shell
            + ElemShapes.volume
            + ElemShapes.beam
            + ElemShapes.springs
            + ElemShapes.masses
            + ElemShapes.other
        ):
            return False
        else:
            return True

    @staticmethod
    def num_nodes(el_name):
        if el_name in ElemShapes.masses + ElemShapes.spring1n:
            return 1
        elif el_name in ElemShapes.bm2 + ElemShapes.spring2n + ElemShapes.other2n:
            return 2
        elif el_name in ElemShapes.tri:
            return 3
        elif el_name in ElemShapes.quad + ElemShapes.tetrahedron:
            return 4
        elif el_name in ElemShapes.pyramid5:
            return 5
        elif el_name in ElemShapes.quad6 + ElemShapes.prism6:
            return 6
        elif el_name in ElemShapes.quad8 + ElemShapes.cube8:
            return 8
        elif el_name in ElemShapes.tetrahedron10:
            return 10
        elif el_name in ElemShapes.prism15:
            return 15
        elif el_name in ElemShapes.cube20:
            return 20
        elif el_name in ElemShapes.cube27:
            return 27
        else:
            raise ValueError(f'element type "{el_name}" is not yet supported')

    def __init__(self, el_type, nodes):
        self.type = el_type.upper()
        if ElemShapes.is_valid_elem(el_type) is False:
            raise ValueError(f'Currently unsupported element type "{el_type}".')

        num_nodes = ElemShapes.num_nodes(self.type)
        if len(nodes) != num_nodes:
            raise ValueError(f'Number of passed nodes "{len(nodes)}" does not match expected "{num_nodes}" ')

        self.nodes = nodes
        self._edges = None

    @property
    def edges(self):
        from ada import Node

        if self.edges_seq is None:
            raise ValueError(f'Element type "{self.type}" is missing element node descriptions')
        if self._edges is None:
            if type(self.nodes[0]) is Node:
                self._edges = [self.nodes[e].p for ed_seq in self.edges_seq for e in ed_seq]
            else:
                self._edges = [self.nodes[e] for ed_seq in self.edges_seq for e in ed_seq]

            return self._edges

        else:
            return self._edges

    @property
    def faces(self):
        if self.type.upper() in ElemShapes.volume:
            faces = self._cube_faces_global
        elif self.type.upper() in ElemShapes.shell:
            faces = self._shell_faces
        else:
            raise ValueError(f'element type "{self.type}" is yet to be included')

        if faces is None:
            return None

        new_n = [[self.nodes[n[0]], self.nodes[n[1]], self.nodes[n[2]]] for n in faces]
        return new_n

    @property
    def _cube_faces_global(self):
        if self._cube_faces is None:
            return None
        new_n = []
        for n in self._cube_faces:
            new_n.append([self.nodes[n[0]], self.nodes[n[1]], self.nodes[n[2]]])
        return new_n

    @property
    def _cube_edges_global(self):
        if self._volume_edges is None:
            return None
        new_n = []
        for n in self._volume_edges:
            new_n.append([self.nodes[n[0]], self.nodes[n[1]]])
        return new_n

    @property
    def edges_seq(self):
        """

        :return:
        :rtype: numpy.ndarray
        """
        if self.type in self.volume:
            edges = self._volume_edges
        elif self.type in self.shell:
            edges = self._shell_edges
        elif self.type in self.beam:
            edges = self._beam_edges
        elif self.type in self.masses + self.spring1n:
            # These are point elements and have no edges
            return None
        elif self.type in self.spring2n + self.connectors:
            # To be implemented
            return None
        else:
            raise ValueError(f'Element type "{self.type}" is yet to be included')

        if edges is not None:
            return np.array(edges)
        else:
            return None

    @property
    def spring_edges(self):
        if self.type not in self.springs:
            return None
        springs = dict(SPRING2=[[0, 1]])
        return springs[self.type]

    @property
    def _beam_edges(self):
        r"""
                Line:                 Line3:          Line4:

              v
              ^
              |
              |
        0-----+-----1 --> u   0----2----1     0---2---3---1

        """
        if self.type not in self.beam:
            logging.error("A call was made to beam edges even though type is not beam")
            return None
        if self.type in ["B31", "SPRING2"]:
            return [[0, 1]]
        elif self.type == "B32":
            return [[0, 2, 1]]
        else:
            raise ValueError(f'Elem type "{self.type}" is not yet supported')

    @property
    def _shell_edges(self):
        r"""Quadrangle:            Quadrangle8:            Quadrangle9:

      v
      ^
      |
3-----------2          3-----6-----2           3-----6-----2
|     |     |          |           |           |           |
|     |     |          |           |           |           |
|     +---- | --> u    7           5           7     8     5
|           |          |           |           |           |
|           |          |           |           |           |
0-----------1          0-----4-----1           0-----4-----1

Triangle:               Triangle6:          Triangle9/10:          Triangle12/15:

v
^                                                                   2
|                                                                   | \
2                       2                    2                      9   8
|`\                     |`\                  | \                    |     \
|  `\                   |  `\                7   6                 10 (14)  7
|    `\                 5    `4              |     \                |         \
|      `\               |      `\            8  (9)  5             11 (12) (13) 6
|        `\             |        `\          |         \            |             \
0----------1 --> u      0-----3----1         0---3---4---1          0---3---4---5---1
"""
        if self.type not in self.shell:
            return None

        if self.type in ["S4", "S4R", "R3D4"]:
            return [[0, 1], [1, 2], [2, 3], [3, 0]]
        elif self.type in ["S3", "S3R"]:
            return [[0, 1], [1, 2], [2, 0]]
        else:
            raise ValueError(f'Elem type "{self.type}" is not yet supported')

    @property
    def _shell_faces(self):
        if self.type not in self.shell:
            return None

        if self.type.upper() in ["S4", "S4R"]:
            return [[0, 1, 2], [0, 2, 3]]
        elif self.type.upper() in ["S3", "S3R"]:
            return [[0, 1, 2]]
        else:
            logging.error(f'element type "{self.type}" is yet to be included')

    @property
    def _volume_edges(self):
        r"""
        Hexahedron:             Hexahedron20:          Hexahedron27:

       v
3----------2            3----13----2           3----13----2
|\     ^   |\           |\         |\          |\         |\
| \    |   | \          | 15       | 14        |15    24  | 14
|  \   |   |  \         9  \       11 \        9  \ 20    11 \
|   7------+---6        |   7----19+---6       |   7----19+---6
|   |  +-- |-- | -> u   |   |      |   |       |22 |  26  | 23|
0---+---\--1   |        0---+-8----1   |       0---+-8----1   |
 \  |    \  \  |         \  17      \  18       \ 17    25 \  18
  \ |     \  \ |         10 |        12|        10 |  21    12|
   \|      w  \|           \|         \|          \|         \|
    4----------5            4----16----5           4----16----5
        Tetrahedron:                          Tetrahedron10:

                   v
                 .
               ,/
              /
           2                                     2
         ,/|`\                                 ,/|`\
       ,/  |  `\                             ,/  |  `\
     ,/    '.   `\                         ,6    '.   `5
   ,/       |     `\                     ,/       8     `\
 ,/         |       `\                 ,/         |       `\
0-----------'.--------1 --> u         0--------4--'.--------1
 `\.         |      ,/                 `\.         |      ,/
    `\.      |    ,/                      `\.      |    ,9
       `\.   '. ,/                           `7.   '. ,/
          `\. |/                                `\. |/
             `3                                    `3
                `\.
                   ` w
        """
        if self.type not in self.volume:
            return None
        if self.type in ["C3D8", "C3D8R", "C3D8H"]:
            return [
                [0, 1],
                [1, 2],
                [2, 3],
                [3, 0],
                [0, 4],
                [4, 7],
                [7, 3],
                [3, 0],
                [4, 5],
                [5, 6],
                [6, 7],
                [1, 5],
                [2, 6],
            ]
        elif self.type in ["C3D4"]:
            return [(0, 1), (1, 3), (3, 0), (0, 2), (2, 3), (1, 2)]
        elif self.type in ["C3D5"]:
            return [(0, 1), (1, 2), (2, 3), (3, 0), (0, 4), (1, 4), (2, 4), (3, 4)]
        elif self.type == "C3D10":
            return [
                [0, 7],
                [7, 3],
                [3, 9],
                [9, 1],
                [1, 4],
                [4, 0],
                [0, 6],
                [6, 2],
                [2, 5],
                [5, 1],
                [3, 8],
                [8, 2],
            ]
        elif self.type in ["C3D20", "C3D20R"]:
            # Abaqus
            return [
                (0, 8),
                (8, 1),
                (1, 17),
                (17, 5),
                (5, 12),
                (12, 4),
                (4, 15),
                (15, 7),
                (0, 11),
                (11, 3),
                (1, 9),
                (9, 2),
                (2, 18),
                (18, 6),
                (6, 14),
                (14, 7),
                (3, 19),
                (19, 7),
                (2, 10),
                (10, 3),
                (5, 13),
                (13, 6),
                (0, 16),
                (16, 4),
            ]
            # gmsh
            # return [
            #     (0, 9), (9, 3), (3, 13), (13, 2), (2, 11), (11, 1), (1, 8), (8, 0),
            #     (0, 10), (10, 4), (4, 17), (17, 7), (7, 15), (15, 3), (3, 9), (9, 0),
            #     (4, 16), (16, 5), (5, 10), (10, 6), (6, 19), (19, 7), (7, 17), (17, 4),
            #     (5, 12), (12, 1), (1, 11), (11, 2), (2, 14), (14, 6), (6, 18), (18, 5),
            #     (0,8), (8,1), (3,13), (13,2)
            # ]
        else:
            print("Element type {} is currently not supported".format(self.type))

    @property
    def _cube_faces(self):
        if self.type not in self.volume:
            return None
        if self.type.upper() == "C3D8":
            return [
                [0, 2, 3],
                [0, 1, 3],
                [0, 4, 7],
                [0, 7, 3],
                [0, 4, 5],
                [0, 5, 1],
                [2, 7, 6],
                [2, 3, 7],
                [5, 6, 7],
                [5, 7, 4],
                [5, 2, 1],
                [5, 6, 2],
            ]
        elif self.type.upper() == "C3D10":
            return [[0, 2, 3], [0, 1, 2], [1, 2, 3], [0, 2, 3]]
        else:
            logging.error(f"Element type {self.type} is currently not supported for visualization")


class FEM(FemBase):
    """
    A FEM representation of its parent Part

    :param name: Name of analysis model
    :param parent: Part object
    :param metadata: Attached metadata
    """

    def __init__(self, name=None, parent=None, metadata=None):
        from ada.core.containers import Nodes

        metadata = metadata if metadata is not None else dict()
        metadata["sensor_data"] = dict()
        metadata["info"] = dict()
        super().__init__(name, metadata, parent)
        self._nodes = Nodes(parent=self)
        self._elements = FemElements(fem_obj=self)
        self._sets = FemSets(fem_obj=self)
        self._sections = FemSections(fem_obj=self)
        self._bcs = []
        self._masses = dict()
        self._constraints = []
        self._surfaces = dict()
        self._amplitudes = dict()
        self._steps = list()
        self._connectors = dict()
        self._connector_sections = dict()
        self._springs = dict()
        self._intprops = dict()
        self._interactions = dict()
        self._sensors = dict()
        self._predefined_fields = dict()
        self._subroutine = None
        self._initial_state = None
        self._lcsys = dict()

    def add_elem(self, elem):
        """

        :param elem:
        :type elem: Elem
        :return:
        """
        elem.parent = self
        self.elements.add(elem)

    def add_section(self, section):
        """

        :param section:
        :type section: FemSection
        """
        section.parent = self
        self.sections.add(section)

    def add_bc(self, bc):
        """
        Adds a BC to the assembly

        :param bc: Bc object
        :type bc: Bc
        """
        if bc.name in [b.name for b in self._bcs]:
            raise Exception('BC with name "{bc_id}" already exists'.format(bc_id=bc.name))
        bc.parent = self
        if bc.fem_set.parent is None:
            # TODO: look over this implementation. Is this okay?
            logging.error("Bc FemSet has no parent. Adding to self")
            self.sets.add(bc.fem_set)

        self._bcs.append(bc)

    def add_mass(self, mass):
        """

        :param mass:
        :type mass: Mass
        """
        mass.parent = self
        self._masses[mass.name] = mass

    def add_set(
        self,
        fem_set,
        ids=None,
        p=None,
        vol_box=None,
        vol_cyl=None,
        single_member=False,
        tol=1e-4,
    ):
        """
        Simple method that creates a set string based on a set name, node or element ids and adds it to the assembly str

        :param fem_set: A fem set object
        :param ids: List of integers
        :param p: Single point (x,y,z)
        :param vol_box: Search by quadratic volume. Where p is (xmin, ymin, zmin) and vol_box is (xmax, ymax, zmax)
        :param vol_cyl: Search by cylindrical volume. Used together with p to find
                        nodes within cylinder inputted by [radius, height, thickness]
        :param single_member: Set True if you wish to keep only a single member
        :param tol: Point Tolerances. Default is 1e-4
        :type fem_set: FemSet
        """
        if ids is not None:
            fem_set.add_members(ids)

        fem_set.parent = self

        def append_members(nodelist):
            if single_member is True:
                fem_set.add_members([nodelist[0]])
            else:
                fem_set.add_members(nodelist)

        if fem_set.type == "nset":
            if p is not None or vol_box is not None or vol_cyl is not None:
                nodes = self.nodes.get_by_volume(p, vol_box, vol_cyl, tol)
                if len(nodes) == 0 and self.parent is not None:
                    assembly = self.parent.get_assembly()
                    list_of_ps = assembly.get_all_subparts() + [assembly]
                    for part in list_of_ps:
                        nodes = part.fem.nodes.get_by_volume(p, vol_box, vol_cyl, tol)
                        if len(nodes) > 0:
                            fem_set.parent = part.fem
                            append_members(nodes)
                            part.fem.add_set(fem_set)
                            return fem_set

                    raise Exception(f'No nodes found for fem set "{fem_set.name}"')
                elif nodes is not None and len(nodes) > 0:
                    append_members(nodes)
                else:
                    raise Exception(f'No nodes found for femset "{fem_set.name}"')

        self._sets.add(fem_set)
        return fem_set

    def add_step(self, step):
        """
        Adds a analysis step to the assembly

        :param step: Step Object
        :type step: Step
        """
        if len(self._steps) > 0:
            if self._steps[-1].type != "eigenfrequency" and step.type == "complex_eig":
                raise Exception(
                    "complex eigenfrequency analysis step needs to follow eigenfrequency step. Check your input"
                )
        step.parent = self
        self._steps.append(step)

        return step

    def add_interaction_property(self, int_prop):
        """

        :param int_prop:
        :type int_prop: InteractionProperty
        """
        int_prop.parent = self
        self._intprops[int_prop.name] = int_prop

    def add_interaction(self, interaction):
        """

        :param interaction:
        :type interaction: Interaction
        """
        interaction.parent = self
        self._interactions[interaction.name] = interaction

    def add_constraint(self, constraint):
        """
        Method for adding a tie or coupling in the assembly

        :param constraint: Name of connection
        :type constraint: Constraint
        """
        constraint.parent = self
        self._constraints.append(constraint)

    def add_lcsys(self, lcsys):
        """

        :param lcsys:
        :type lcsys: ada.fem.Csys
        :return:
        """
        if lcsys.name in self._lcsys.keys():
            raise ValueError("Local Coordinate system cannot have duplicate name")
        lcsys.parent = self
        self._lcsys[lcsys.name] = lcsys

    def add_connector_section(self, connector_section):
        """
        Adds a connector section to the assembly

        :param connector_section:
        :type connector_section: ConnectorSection
        """
        connector_section.parent = self
        self._connector_sections[connector_section.name] = connector_section

    def add_connector(self, connector):
        """
        Add a connector to the assembly using the :func:`~abaqus.utils.add_connector_elem`

        :param connector: Name of connector
        :type connector: Connector

        """
        connector.parent = self
        self._connectors[connector.name] = connector
        connector.csys.parent = self
        self.elements.add(connector)
        self.add_set(FemSet(name=connector.name, members=[connector.id], set_type="elset"))

    def add_rp(self, name, node):
        """
        Adds a reference point in assembly

        :param node: Node
        :param name: Creates a set with by specifying a name
        :type node: ada.Node
        """
        node.parent = self
        self.nodes.add(node)
        fem_set = self.add_set(FemSet(name, [node], "nset"))
        return node, fem_set

    def add_surface(self, surface):
        """

        :param surface:
        :type surface: Surface
        """
        surface.parent = self
        self._surfaces[surface.name] = surface

    def add_amplitude(self, amplitude):
        """
        Adds amplitude data history to the assembly


        :param amplitude: Amplitude object
        :type amplitude: Amplitude
        """
        amplitude.parent = self
        self._amplitudes[amplitude.name] = amplitude

    def add_sensor(self, name, point, comment, tol=1e-2):
        """

        :param name: Name of coordinate set
        :param point: Sensor Coordinate
        :param comment: Comment
        :param tol:
        """
        fem_set = FemSet(name, [], "nset", metadata=dict(comment=comment))
        self.add_set(fem_set, p=point, tol=tol, single_member=True)
        if name in self._sensors.keys():
            raise Exception("{} exists in sensor sets and will be overwritten. Please change name.".format(name))

        self._sensors[name] = fem_set
        # self._cad[name].add_shape(ada.cad.utils.make_sphere(point, tol), colour='red', transparency=0.5)

    def add_predefined_field(self, pre_field):
        """

        :type pre_field: PredefinedField
        """
        pre_field.parent = self

        self._predefined_fields[pre_field.name] = pre_field

    def add_spring(self, spring):
        """

        :param spring:
        :type spring: Spring
        """

        # self.elements.add(spring)

        if spring.fem_set.parent is None:
            self.sets.add(spring.fem_set)
        self._springs[spring.name] = spring

    def convert_ecc_to_mpc(self):
        """

        Converts beam offsets to MPC constraints
        """
        from ada import Node
        from ada.core.utils import vector_length

        edited_nodes = dict()
        tol = 1e-3

        def build_mpc(fs):
            """

            :param fs:
            :type fs: FemSection
            :return:
            """
            if fs.offset is None or fs.type != "beam":
                return
            elem = fs.elset.members[0]
            for n_old, ecc in fs.offset:
                i = elem.nodes.index(n_old)
                if n_old.id in edited_nodes.keys():
                    n_new = edited_nodes[n_old.id]
                    mat = np.eye(3)
                    new_p = np.dot(mat, ecc) + n_old.p
                    n_new_ = Node(new_p, parent=elem.parent)
                    if vector_length(n_new_.p - n_new.p) > tol:
                        elem.parent.nodes.add(n_new_, allow_coincident=True)
                        m_set = FemSet(f"el{elem.id}_mpc{i + 1}_m", [n_new_], "nset")
                        s_set = FemSet(f"el{elem.id}_mpc{i + 1}_s", [n_old], "nset")
                        c = Constraint(
                            f"el{elem.id}_mpc{i + 1}_co",
                            "mpc",
                            m_set,
                            s_set,
                            mpc_type="Beam",
                            parent=elem.parent,
                        )
                        elem.parent.add_constraint(c)
                        elem.nodes[i] = n_new_
                        edited_nodes[n_old.id] = n_new_

                    else:
                        elem.nodes[i] = n_new
                        edited_nodes[n_old.id] = n_new
                else:
                    mat = np.eye(3)
                    new_p = np.dot(mat, ecc) + n_old.p
                    n_new = Node(new_p, parent=elem.parent)
                    elem.parent.nodes.add(n_new, allow_coincident=True)
                    m_set = FemSet(f"el{elem.id}_mpc{i + 1}_m", [n_new], "nset")
                    s_set = FemSet(f"el{elem.id}_mpc{i + 1}_s", [n_old], "nset")
                    c = Constraint(
                        f"el{elem.id}_mpc{i + 1}_co",
                        "mpc",
                        m_set,
                        s_set,
                        mpc_type="Beam",
                        parent=elem.parent,
                    )
                    elem.parent.add_constraint(c)

                    elem.nodes[i] = n_new
                    edited_nodes[n_old.id] = n_new

        list(map(build_mpc, filter(lambda x: x.offset is not None, self.sections)))

    def convert_hinges_2_couplings(self):
        """
        Convert beam hinges to coupling constraints
        """
        from ada import Node

        def converthinges(fs):
            """

            :param fs:
            :type fs: ada.fem.FemSection
            """
            if fs.hinges is None or fs.type != "beam":
                return
            elem = fs.elset.members[0]
            assert isinstance(elem, Elem)

            for n, d, csys in fs.hinges:
                n2 = Node(n.p, None, parent=elem.parent)
                elem.parent.nodes.add(n2, allow_coincident=True)
                i = elem.nodes.index(n)
                elem.nodes[i] = n2
                s_set = FemSet(f"el{elem.id}_hinge{i + 1}_s", [n], "nset")
                m_set = FemSet(f"el{elem.id}_hinge{i + 1}_m", [n2], "nset")
                elem.parent.add_set(m_set)
                elem.parent.add_set(s_set)
                c = Constraint(
                    f"el{elem.id}_hinge{i + 1}_co",
                    "coupling",
                    m_set,
                    s_set,
                    d,
                    csys=csys,
                )
                elem.parent.add_constraint(c)

        list(map(converthinges, filter(lambda x: x.hinges is not None, self.sections)))

    @property
    def parent(self):
        """

        :rtype: ada.Part
        """
        return self._parent

    @parent.setter
    def parent(self, value):
        from ada import Part

        if issubclass(type(value), Part) is False and value is not None:
            raise ValueError()
        self._parent = value

    @property
    def nodes(self):
        """

        :return:
        :rtype: ada.core.containers.Nodes
        """
        return self._nodes

    @property
    def elements(self):
        """

        :return:
        :rtype: FemElements
        """
        return self._elements

    @property
    def sections(self):
        """

        :return:
        :rtype: FemSections
        """
        return self._sections

    @property
    def bcs(self):
        return self._bcs

    @property
    def constraints(self):
        return self._constraints

    @property
    def instance_name(self):
        return self._name if self._name is not None else f"{self.parent.name}-1"

    @property
    def sets(self):
        """

        :return:
        :rtype: ada.fem.containers.FemSets
        """
        return self._sets

    @property
    def nsets(self):
        return self.sets.nodes

    @property
    def elsets(self):
        return self.sets.elements

    @property
    def masses(self):
        return self._masses

    @property
    def interactions(self):
        """

        :rtype: dict
        """
        return self._interactions

    @property
    def intprops(self):
        return self._intprops

    @property
    def steps(self):
        """

        :return:
        :rtype: list
        """
        return self._steps

    @property
    def surfaces(self):
        """

        :rtype: dict
        """
        return self._surfaces

    @property
    def connectors(self):
        return self._connectors

    @property
    def connector_sections(self):
        return self._connector_sections

    @property
    def amplitudes(self):
        return self._amplitudes

    @property
    def sensors(self):
        return self._sensors

    @property
    def predefined_fields(self):
        return self._predefined_fields

    @property
    def initial_state(self):
        """

        :return:
        :rtype: PredefinedField
        """
        return self._initial_state

    @initial_state.setter
    def initial_state(self, value):
        self._initial_state = value

    @property
    def springs(self):
        return self._springs

    @property
    def lcsys(self):
        return self._lcsys

    def __repr__(self):
        return f"FEM({self.name}, Elements: {len(self.elements)}, Nodes: {len(self.nodes)})"


class FemSection(FemBase):
    """

    :param name:
    :param sec_type:
    :param elset:
    :param material:
    :param section:
    :param thickness:
    :param int_points:
    :type elset: FemSet
    :type material: ada.Material
    :type section: ada.Section
    """

    def __init__(
        self,
        name,
        sec_type,
        elset,
        material,
        section=None,
        local_z=None,
        local_y=None,
        thickness=None,
        int_points=5,
        offset=None,
        hinges=None,
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        self._sec_type = sec_type
        if sec_type is None:
            raise ValueError("Section type cannot be None")
        self._elset = elset
        self._material = material
        self._section = section
        self._local_z = local_z
        self._local_y = local_y
        self._thickness = thickness
        self._int_points = int_points
        self._offset = offset
        self._hinges = hinges

    def link_elements(self):
        def link_elem(el):
            """

            :param el:
            :type el: Elem
            """
            el.fem_sec = self

        list(map(link_elem, self.elset.members))

    def get_offset_coords(self):
        """

        :param self:
        :type self: FemSection
        :return:
        """

        elem = self.elset.members[0]
        nodes = [n.p for n in elem.nodes]
        if self.offset is None:
            return nodes

        for n_old, ecc in self.offset:
            mat = np.eye(3)
            new_p = np.dot(mat, ecc) + n_old.p
            i = elem.nodes.index(n_old)
            nodes[i] = new_p

        return nodes

    @property
    def type(self):
        return self._sec_type

    @property
    def elset(self):
        return self._elset

    @property
    def local_z(self):
        """

        :return: Local Z describes the up vector of the cross section
        :rtype: list
        """
        from ada.core.utils import roundoff, unit_vector, vector_length

        if self._local_z is None:
            if self.type == "beam":
                n1, n2 = self.elset.members[0].nodes[0], self.elset.members[0].nodes[-1]
                v = n2.p - n1.p
                if vector_length(v) == 0.0:
                    logging.error(f"Element {self.elset.members[0].id} has zero length")
                    xvec = [1, 0, 0]
                else:
                    xvec = unit_vector(v)

                crossed = np.cross(xvec, self.local_y)
                ma = max(abs(crossed))
                self._local_z = tuple([roundoff(x / ma, 3) for x in crossed])
            else:
                from ada.core.utils import normal_to_points_in_plane

                self._local_z = normal_to_points_in_plane([n.p for n in self.elset.members[0].nodes])
                # raise NotImplementedError("Local Z is not implemented for shell elements, yet.")
        return self._local_z

    @property
    def local_y(self):
        """



        :return: Local Z describes the up vector of the cross section
        :rtype: list
        """
        from ada.core.utils import roundoff, unit_vector, vector_length

        if self._local_y is None:
            if self.type in ("beam", "shell"):
                n1, n2 = self.elset.members[0].nodes[0], self.elset.members[0].nodes[-1]
                v = n2.p - n1.p
                if vector_length(v) == 0.0:
                    xvec = [1, 0, 0]
                else:
                    xvec = unit_vector(v)
                # See https://en.wikipedia.org/wiki/Cross_product#Coordinate_notation for order of cross product
                crossed = np.cross(self.local_z, xvec)
                ma = max(abs(crossed))
                self._local_y = tuple([roundoff(x / ma, 3) for x in crossed])
            else:
                raise NotImplementedError("Local Y is not implemented for solid elements.")
        return self._local_y

    @property
    def local_x(self):
        if self.type == "beam":
            from ada.core.utils import unit_vector

            el = self.elset.members[0]
            return unit_vector(el.nodes[-1].p - el.nodes[0].p)
        else:
            logging.error(f"X-vector not defined for {self.type}")

    @property
    def csys(self):
        return [self.local_x, self.local_y, self.local_z]

    @property
    def section(self):
        """

        :return:
        :rtype: ada.Section
        """
        return self._section

    @property
    def material(self):
        """

        :return:
        :rtype: ada.Material
        """
        return self._material

    @property
    def thickness(self):
        return self._thickness

    @property
    def int_points(self):
        return self._int_points

    @property
    def offset(self):
        return self._offset

    @property
    def hinges(self):
        return self._hinges

    def __eq__(self, other):
        for key, val in self.__dict__.items():
            if "parent" in key:
                continue
            # Re-evaluate the elset exemption. Maybe this should raise False based on the fem set only?
            # if 'elset' in key:
            #     for m in self.__dict__[key].members:
            #         if m.type != other.__dict__[key].members[0].type:
            #             return False
            if other.__dict__[key] != val:
                return False

        return True

    def __repr__(self):
        return (
            f'FemSection({self.type} - name: "{self.name}", sec: "{self.section}", '
            f'mat: "{self.material}",  elset: "{self.elset}")'
        )


class FemSet(FemBase):
    """

    :param name: Name of Set
    :param members: Set Members
    :param set_type: Type of set (either 'nset' or 'elset')
    :param metadata: Metadata for object
    :param parent: Parent object
    """

    _valid_types = ["nset", "elset"]

    def __init__(self, name, members, set_type, metadata=None, parent=None):
        super().__init__(name, metadata, parent)
        self._set_type = set_type
        if self.type not in FemSet._valid_types:
            raise ValueError(f'set type "{set_type}" is not valid')
        self._members = members

    def __len__(self):
        return len(self._members)

    def __contains__(self, item):
        return item.id in self._members

    def __getitem__(self, index):
        return self._members[index]

    def __add__(self, other):
        """

        :param other:
        :type other: FemSet
        :return:
        """
        self.add_members(other.members)
        return self

    def add_members(self, members):
        """

        :param members:
        :type members: list
        """

        self._members += members

    @property
    def type(self):
        """

        :return: Type of set
        """
        return self._set_type.lower()

    @property
    def members(self):
        """

        :return: Members of set
        """
        return self._members

    @property
    def instance_num(self):
        """

        :return:
        """
        if self.on_assembly_level is True:
            return ",".join([f"{m}" for m in self.members])
        else:
            return ",".join(["{}.{}".format(self.parent.instance_name, m) for m in self.members])

    def __repr__(self):
        return f'FemSet({self.name}, type: "{self.type}", members: "{len(self.members)}")'


class Amplitude(FemBase):
    """

    :param name:
    :param x:
    :param y:
    :param smooth:
    :param metadata:
    :param parent:
    """

    def __init__(self, name, x, y, smooth=None, metadata=None, parent=None):
        super().__init__(name, metadata, parent)
        self._x = x
        self._y = y
        self._smooth = smooth

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def smooth(self):
        return self._smooth


class Surface(FemBase):
    """
    Documentation

        https://abaqus-docs.mit.edu/2017/English/SIMACAEKEYRefMap/simakey-r-surface.htm#simakey-r-surface__simakey-r-surface-s-datadesc5


    Parameters.

    :param name: Unique name of surface
    :param surf_type: Type of surface
    :param fem_set:
    :param weight_factor:
    :param id_refs: Explicitly defined by list of tuple [(elid/nid,spos), ..]
    :param parent:
    :param metadata:
    :type fem_set: FemSet
    """

    _valid_surf_types = ["ELEMENT", "NODE"]

    def __init__(
        self,
        name,
        surf_type,
        fem_set,
        weight_factor=None,
        face_id_label=None,
        id_refs=None,
        parent=None,
        metadata=None,
    ):
        super().__init__(name, metadata, parent)

        if surf_type not in self._valid_surf_types:
            raise ValueError(f'Surface type "{surf_type}" is currently not supported\\implemented')

        self._surf_type = surf_type
        self._fem_set = fem_set
        self._weight_factor = weight_factor
        self._face_id_label = face_id_label
        self._id_refs = id_refs

    @property
    def type(self):
        return self._surf_type

    @property
    def fem_set(self):
        return self._fem_set

    @property
    def weight_factor(self):
        return self._weight_factor

    @property
    def face_id_label(self):
        return self._face_id_label

    @property
    def id_refs(self):
        return self._id_refs


class Elem(FemBase):
    """
    Node numbering of elements is based on GMSH doc here http://gmsh.info/doc/texinfo/gmsh.html#Node-ordering

    :param el_id:
    :param nodes:
    :param el_type:
    :param elset:
    :param fem_sec:
    :param mass_props:
    :param parent:
    :param metadata:
    :type fem_sec: FemSection
    """

    def __init__(
        self,
        el_id,
        nodes,
        el_type,
        elset=None,
        fem_sec=None,
        mass_props=None,
        parent=None,
        metadata=None,
    ):
        from ada import Node

        super().__init__(el_id, metadata, parent)
        self.type = el_type.upper()
        self._el_id = el_id

        self._shape = ElemShapes(self.type, nodes)

        if type(nodes[0]) is Node:
            for node in nodes:
                node.refs.append(self)

        self._nodes = nodes
        self._elset = elset
        self._fem_sec = fem_sec
        self._mass_props = mass_props
        self._refs = []

    @property
    def type(self):
        """

        :return: Element type.
        """
        return self._el_type

    @type.setter
    def type(self, value):
        if ElemShapes.is_valid_elem(value) is False:
            raise ValueError(f'Currently unsupported element type "{value}".')
        self._el_type = value.upper()

    @property
    def name(self):
        """

        :return:
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def id(self):
        return self._el_id

    @id.setter
    def id(self, value):
        if type(value) not in (np.int32, int, np.uint64) and issubclass(type(self), Connector) is False:
            raise ValueError(f'Element name type "{type(value)}" must be numeric')
        self._el_id = value

    @property
    def nodes(self):
        return self._nodes

    @property
    def elset(self):
        return self._elset

    @property
    def fem_sec(self):
        """

        :return:
        :rtype: ada.fem.FemSection
        """
        return self._fem_sec

    @fem_sec.setter
    def fem_sec(self, value):
        self._fem_sec = value

    @property
    def mass_props(self):
        """

        :return:
        :rtype: ada.fem.Mass
        """
        return self._mass_props

    @mass_props.setter
    def mass_props(self, value):
        self._mass_props = value

    @property
    def shape(self):
        """

        :return:
        :rtype: ada.fem.ElemShapes
        """
        return self._shape

    @property
    def refs(self):
        return self._refs

    def __repr__(self):
        return f'Elem(ID: {self._el_id}, Type: {self.type}, NodeIds: "{self.nodes}")'


class Connector(Elem):
    """
    A Connector Element

    :param name:
    :param con_type:
    :param con_sec:
    :param el_id:
    :param n1:
    :param n2:
    :param csys:

    :type n1: ada.Node
    :type n2: ada.Node
    :type con_sec: ConnectorSection
    :type csys: Csys
    """

    def __init__(
        self,
        name,
        el_id,
        n1,
        n2,
        con_type,
        con_sec,
        preload=None,
        csys=None,
        metadata=None,
        parent=None,
    ):
        from ada import Node

        if type(n1) is not Node or type(n2) is not Node:
            raise ValueError("Connector Start\\end must be nodes")
        super(Connector, self).__init__(el_id, [n1, n2], "CONNECTOR")
        super(Elem, self).__init__(name, metadata, parent)
        self._n1 = n1
        self._n2 = n2
        self._con_type = con_type
        self._con_sec = con_sec
        self._preload = preload
        self._csys = csys if csys is not None else Csys(f"{name}_csys")

    @property
    def con_type(self):
        return self._con_type

    @property
    def con_sec(self):
        """

        :return:
        :rtype: ConnectorSection
        """
        return self._con_sec

    @property
    def n1(self):
        """

        :return:
        :rtype: ada.Node
        """
        return self._n1

    @property
    def n2(self):
        """

        :return:
        :rtype: ada.Node
        """
        return self._n2

    @property
    def csys(self):
        """

        :return:
        :rtype: Csys
        """
        return self._csys

    def __repr__(self):
        return f'ConnectorElem(ID: {self.id}, Type: {self.type}, End1: "{self.n1}", End2: "{self.n2}")'


class Spring(Elem):
    """

    :param name:
    :param el_id:
    :param el_type:
    :param stiff:
    :param n1:
    :param n2:
    :param metadata:
    :param parent:
    """

    def __init__(self, name, el_id, el_type, stiff, n1, n2=None, metadata=None, parent=None):
        nids = [n1]
        if n2 is not None:
            nids += [n2]
        super(Spring, self).__init__(el_id, nids, el_type)
        super(Elem, self).__init__(name, metadata, parent)
        self._stiff = stiff
        self._n1 = n1
        self._n2 = n2
        self._fem_set = FemSet(self.name + "_set", [el_id], "elset")
        if self.parent is not None:
            self.parent.sets.add(self._fem_set)

    @property
    def fem_set(self):
        return self._fem_set

    @property
    def stiff(self):
        return self._stiff

    def __repr__(self):
        return f'Spring("{self.name}", type="{self._stiff}")'


class Csys(FemBase):
    """

    :param name:
    :param nodes:
    :param coords:
    :param metadata:
    :param parent:
    """

    _valid_systems = ["RECTANGULAR"]  # , 'CYLINDRICAL', 'SPHERICAL', 'Z RECTANGULAR', 'USER']
    _valid_defs = ["COORDINATES", "NODES"]  # ,'OFFSET TO NODES'

    def __init__(
        self,
        name,
        definition="COORDINATES",
        system="RECTANGULAR",
        nodes=None,
        coords=None,
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        self._definition = definition
        self._system = system
        self._nodes = nodes
        self._coords = coords

    @property
    def definition(self):
        return self._definition

    @property
    def system(self):
        return self._system

    @property
    def nodes(self):
        return self._nodes

    @property
    def coords(self):
        return self._coords

    def __repr__(self):
        content_map = dict(COORDINATES=self.coords, NODES=self.nodes)
        return f'Csys("{self.name}", "{self.definition}", {content_map[self.definition]})'


class ConnectorSection(FemBase):
    """
    A Connector Section


    :param name:
    :param elastic_comp:
    :param damping_comp:
    :param plastic_comp:
    :param rigid_dofs:
    :param soft_elastic_dofs:
    """

    def __init__(
        self,
        name,
        elastic_comp,
        damping_comp,
        plastic_comp=None,
        rigid_dofs=None,
        soft_elastic_dofs=None,
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        self._elastic_comp = elastic_comp if elastic_comp is not None else []
        self._damping_comp = damping_comp if damping_comp is not None else []
        self._plastic_comp = plastic_comp
        self._rigid_dofs = rigid_dofs
        self._soft_elastic_dofs = soft_elastic_dofs

    @property
    def elastic_comp(self):
        return self._elastic_comp

    @property
    def damping_comp(self):
        return self._damping_comp

    @property
    def plastic_comp(self):
        return self._plastic_comp

    @property
    def rigid_dofs(self):
        return self._rigid_dofs

    @property
    def soft_elastic_dofs(self):
        return self._soft_elastic_dofs


class Constraint(FemBase):
    """
    A Constraint

    :param name:
    :param con_type:
    :param m_set:
    :param s_set:
    :param dofs:
    :param pos_tol:
    :param mpc_type:
    :type m_set: FemSet
    :type s_set: FemSet
    """

    def __init__(
        self,
        name,
        con_type,
        m_set,
        s_set,
        dofs=None,
        pos_tol=None,
        mpc_type=None,
        csys=None,
        parent=None,
        metadata=None,
    ):
        super().__init__(name, metadata, parent)
        self._con_type = con_type
        self._m_set = m_set
        self._s_set = s_set
        self._dofs = [1, 2, 3, 4, 5, 6] if dofs is None else dofs
        self._pos_tol = pos_tol
        self._mpc_type = mpc_type
        self._csys = csys

    @property
    def type(self):
        return self._con_type

    @property
    def m_set(self):
        """

        :return:
        :rtype: FemSet
        """
        return self._m_set

    @property
    def s_set(self):
        """

        :return:
        :rtype: FemSet
        """
        return self._s_set

    @property
    def dofs(self):
        return self._dofs

    @property
    def pos_tol(self):
        return self._pos_tol

    @property
    def csys(self):
        """

        :rtype: Csys
        """
        return self._csys

    @property
    def mpc_type(self):
        return self._mpc_type

    def __repr__(self):
        return f'Constraint("{self.type}", m: "{self.m_set.name}", s: "{self.s_set.name}", dofs: "{self.dofs}")'


class InteractionProperty(FemBase):
    """

    :param name:
    :param friction:
    :param pressure_overclosure:
    :param tabular:
    :param metadata:
    :param parent:
    """

    _valid_po = ["HARD", "TABULAR", "PENALTY"]

    def __init__(
        self,
        name,
        friction=None,
        pressure_overclosure="HARD",
        tabular=None,
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        self._friction = friction if friction is not None else 0.0
        self._pressure_overclosure = pressure_overclosure
        if self.pressure_overclosure not in InteractionProperty._valid_po:
            raise ValueError(f'Pressure overclosure type "{pressure_overclosure}" is not supported')
        self._tabular = tabular

    @property
    def friction(self):
        return self._friction

    @property
    def pressure_overclosure(self):
        return self._pressure_overclosure.strip()

    @property
    def tabular(self):
        return self._tabular


class Interaction(FemBase):
    """
    A class representing the physical properties of
    interaction between solid bodies.

    """

    _valid_contact_types = ["SURFACE", "GENERAL"]
    _valid_surface_types = ["SURFACE TO SURFACE"]

    def __init__(
        self,
        name,
        contact_type,
        surf1,
        surf2,
        int_prop,
        constraint=None,
        surface_type="SURFACE TO SURFACE",
        parent=None,
        metadata=None,
    ):
        """

        :param name:
        :param surf1:
        :param surf2:
        :param int_prop:
        :param constraint:
        :param surface_type: Interaction type.
        :type name: str
        :type int_prop: InteractionProperty
        :type constraint: str
        :type surf1: Surface
        :type surf2: Surface
        :type surface_type: str
        :type parent: FEM
        :type metadata: dict
        """
        super().__init__(name, metadata, parent)

        self.type = contact_type
        self.surface_type = surface_type
        self._surf1 = surf1
        self._surf2 = surf2
        self._int_prop = int_prop
        self._constraint = constraint

    @property
    def parent(self):
        """

        :rtype: ada.fem.FEM
        """
        return self._parent

    @parent.setter
    def parent(self, value):
        if type(value) not in (FEM, Step) and value is not None:
            raise ValueError(f'Parent type "{type(value)}" is not supported')
        self._parent = value

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        if value.upper() not in self._valid_contact_types:
            raise ValueError(f'Contact type cannot be "{value}". Must be in {self._valid_contact_types}')
        self._type = value.upper()

    @property
    def surf1(self):
        """

        :rtype: Surface
        """
        return self._surf1

    @property
    def surf2(self):
        """

        :rtype: Surface
        """
        return self._surf2

    @property
    def interaction_property(self):
        """

        :return:
        :rtype: InteractionProperty
        """
        return self._int_prop

    @property
    def constraint(self):
        return self._constraint

    @property
    def surface_type(self):
        return self._surface_type

    @surface_type.setter
    def surface_type(self, value):
        if value not in self._valid_surface_types:
            raise ValueError(f'Surface type cannot be "{value}". Must be in {self._valid_surface_types}')
        self._surface_type = value


class PredefinedField(FemBase):
    """

    :param name:
    :param field_type:
    :param fem_set:
    :param dofs:
    :param magnitude:
    :param metadata:
    :param parent:
    """

    valid_types = ["VELOCITY", "INITIAL STATE"]

    def __init__(
        self,
        name,
        field_type,
        fem_set=None,
        dofs=None,
        magnitude=None,
        initial_state_file=None,
        initial_state_part=None,
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        self.type = field_type
        self._fem_set = fem_set
        self._dofs = dofs
        self._magnitude = magnitude
        self._initial_state_part = initial_state_part
        self._initial_state_file = initial_state_file
        if self.initial_state_file is not None:
            self.initial_state_part.fem.initial_state = self

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        if value.upper() not in PredefinedField.valid_types:
            raise ValueError(f'The field type "{value.upper()}" is currently not supported')
        self._type = value.upper()

    @property
    def fem_set(self):
        """

        :return:
        :rtype: FemSet
        """
        return self._fem_set

    @property
    def dofs(self):
        return self._dofs

    @property
    def magnitude(self):
        return self._magnitude

    @property
    def initial_state_part(self):
        """

        :return:
        :rtype: ada.Part
        """
        return self._initial_state_part

    @property
    def initial_state_file(self):
        return self._initial_state_file


class Bc(FemBase):
    """

    :param name:
    :param fem_set:
    :param dofs:
    :param magnitudes:
    :param bc_type:
    :param amplitude_name:
    :param init_condition: List of tuples [(dof1, magnitude1), (dof2, magnitude2)]
    :type fem_set: FemSet
    """

    _valid_types = [
        "displacement",
        "velocity",
        "connector_displacement",
        "symmetry/antisymmetry/encastre",
        "displacement/rotation",
        "velocity/angular velocity",
    ]

    def __init__(
        self,
        name,
        fem_set,
        dofs,
        magnitudes=None,
        bc_type="displacement",
        amplitude_name=None,
        init_condition=None,
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        self._fem_set = fem_set
        self._dofs = dofs if type(dofs) is list else [dofs]
        if magnitudes is None:
            self._magnitudes = [None] * len(self._dofs)
        else:
            self._magnitudes = magnitudes if type(magnitudes) is list else [magnitudes]
        self.type = bc_type
        self._amplitude_name = amplitude_name
        self._init_condition = init_condition

    def add_init_condition(self, init_condition):
        self._init_condition = init_condition

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        if value.lower() not in self._valid_types:
            raise ValueError(f'BC type "{value}" is not yet supported')
        self._type = value.lower()

    @property
    def fem_set(self):
        """

        :return:
        :rtype: FemSet
        """
        return self._fem_set

    @property
    def dofs(self):
        return self._dofs

    @property
    def magnitudes(self):
        return self._magnitudes

    @property
    def amplitude_name(self):
        return self._amplitude_name

    def __repr__(self):
        return f'Bc("{self.name}", type="{self.type}", dofs={self.dofs}, fem_set="{self.fem_set.name}")'


class Mass(FemBase):
    """

    :param name: Name of set
    :param fem_set: Fem Set (of element or nodal type)
    :type fem_set: FemSet
    :param mass: Mass magnitude
    :param mass_type: Type of mass. See _valid_types. Default is 'MASS'
    :param ptype: Point mass type. Can be None, 'Isotropic' or 'Anisotropic'
    :param units:
    :param metadata:
    :param parent:

    """

    _valid_types = ["MASS", "NONSTRUCTURAL MASS", "ROTARY INERTIA"]
    _valid_ptypes = [None, "ISOTROPIC", "ANISOTROPIC"]

    def __init__(
        self,
        name,
        fem_set,
        mass,
        mass_type=None,
        ptype=None,
        units=None,
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        self._fem_set = fem_set
        if mass is None:
            raise ValueError("Mass cannot be None")
        if type(mass) not in (list, tuple):
            logging.info(f"Mass {type(mass)} converted to list of len=1. Assume equal mass in all 3 transl. DOFs.")
            mass = [mass]
        self._mass = mass
        self._mass_type = mass_type if mass_type is not None else "MASS"
        if self.type not in Mass._valid_types:
            raise ValueError(f'Mass type "{self.type}" is not in list of supported types {self._valid_types}')
        if ptype not in self._valid_ptypes:
            raise ValueError(f'Mass point type "{ptype}" is not in list of supported types {self._valid_ptypes}')
        self.point_mass_type = ptype
        self._units = units

    @property
    def type(self):
        return self._mass_type.upper()

    @property
    def fem_set(self):
        """
        :rtype: FemSet
        """
        return self._fem_set

    @property
    def mass(self):
        if self.point_mass_type is None:
            if self.type == "MASS":
                if type(self._mass) in (list, tuple):
                    raise ValueError("Mass can only be a scalar number for Isotropic mass")
                return float(self._mass[0])
            elif self.type == "NONSTRUCTURAL MASS":
                return self._mass
            else:
                return float(self._mass)
        elif self.point_mass_type == "ISOTROPIC":
            if (len(self._mass) == 1) is False:
                raise ValueError("Mass can only be a scalar number for Isotropic mass")
            return self._mass[0]
        elif self.point_mass_type == "ANISOTROPIC":
            if (len(self._mass) == 3) is False:
                raise ValueError("Mass must be specified for 3 dofs for Anisotropic mass")
            return self._mass
        else:
            raise ValueError(f'Unknown mass input "{self.type}"')

    @property
    def units(self):
        return self._units

    @property
    def point_mass_type(self):
        return self._ptype

    @point_mass_type.setter
    def point_mass_type(self, value):
        self._ptype = value

    def __repr__(self):
        return f"Mass({self.name}, {self.point_mass_type}, [{self.mass}])"


class HistOutput(FemBase):
    """

    :param name: Unique History Output Name
    :param fem_set: Set name associated for history output
    :param set_type:
    :param variables:
    :param int_type: Interval type
    :type name: str
    :type set_type: str
    :type variables: list

    """

    default_hist = [
        "ALLAE",
        "ALLCD",
        "ALLDMD",
        "ALLEE",
        "ALLFD",
        "ALLIE",
        "ALLJD",
        "ALLKE",
        "ALLKL",
        "ALLPD",
        "ALLQB",
        "ALLSD",
        "ALLSE",
        "ALLVD",
        "ALLWK",
        "ETOTAL",
    ]
    _valid_types = ["node", "energy", "contact", "connector"]

    def __init__(
        self,
        name,
        fem_set,
        set_type,
        variables,
        int_value=1,
        int_type="frequency",
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)

        if set_type not in self._valid_types:
            raise ValueError(
                f'set_type "{set_type}" is not yet supported. Currently supported types are "{self._valid_types}"'
            )

        self._fem_set = fem_set
        self._set_type = set_type
        self._variables = variables
        self._int_value = int_value
        self._int_type = int_type

    @property
    def type(self):
        return self._set_type

    @property
    def fem_set(self):
        return self._fem_set

    @property
    def variables(self):
        return self._variables

    @property
    def int_type(self):
        return self._int_type

    @property
    def int_value(self):
        return self._int_value


class FieldOutput(FemBase):
    """
    https://abaqus-docs.mit.edu/2017/English/SIMACAEKEYRefMap/simakey-r-output.htm

    :param name:
    :param nodal:
    :param element:
    :param contact:
    :param int_value: Field output step interval. Default is 1
    :param int_type:
    :param metadata:
    :param parent:
    """

    _valid_fstep_type = ["FREQUENCY", "NUMBER INTERVAL"]
    default_no = ["A", "CF", "RF", "U", "V"]
    default_el = ["LE", "PE", "PEEQ", "PEMAG", "S"]
    default_co = ["CSTRESS", "CDISP", "CFORCE", "CSTATUS"]

    def __init__(
        self,
        name,
        nodal=None,
        element=None,
        contact=None,
        int_value=1,
        int_type="frequency",
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        self._nodal = FieldOutput.default_no if nodal is None else nodal
        self._element = FieldOutput.default_el if element is None else element
        self._contact = FieldOutput.default_co if contact is None else contact
        self._int_value = int_value
        self._int_type = int_type

    @property
    def nodal(self):
        return self._nodal

    @property
    def element(self):
        return self._element

    @property
    def contact(self):
        return self._contact

    @property
    def int_value(self):
        return self._int_value

    @int_value.setter
    def int_value(self, value):
        if value < 0:
            raise ValueError("The interval or frequency value cannot be less than 0")

        self._int_value = value

    @property
    def int_type(self):
        return self._int_type.upper()

    @int_type.setter
    def int_type(self, value):
        if value.upper() not in FieldOutput._valid_fstep_type:
            raise ValueError(f'Field output step type "{value}" is not supported')
        self._int_type = value.upper()


class Step(FemBase):
    """
    A FEM analysis step object

    :param name: Name of step
    :param step_type: Step type: | 'static' | 'eigenfrequency' |  'response_analysis' | 'dynamic' | 'complex_eig' |
    :param nl_geom: Include or ignore the nonlinear effects of large deformations and displacements (default=False)
    :param total_incr: Maximum number of allowed increments
    :param init_incr: Initial increment
    :param total_time: Total step time
    :param min_incr: Minimum allowable increment size
    :param max_incr: Maximum allowable increment size
    :param unsymm: Unsymmetric Matrix storage (default=False)
    :param stabilize: Default=None.
    :param dyn_type: Dynamic analysis type 'TRANSIENT FIDELITY' | 'QUASI-STATIC'
    :param init_accel_calc: Initial acceleration calculation
    :param eigenmodes: Eigenmodes
    :param alpha: Rayleigh Damping for use in Steady State analysis
    :param beta: Rayleigh Damping for use in Steady State analysis
    :param nodeid: Node ID for use in Steady State analysis
    :param fmin: Minimum frequency for use in Steady State analysis
    :param fmax: Maximum frequency for use in Steady State analysis

    :type name: str
    :type step_type: str
    :type nl_geom: bool
    :type total_incr: int
    :type init_incr: float
    :type total_time: float
    :type min_incr: float
    :type max_incr: float
    :type unsymm: bool
    :type stabilize: dict
    :type dyn_type: str
    :type init_accel_calc: bool
    :type eigenmodes: int
    :type alpha: float
    :type beta: float
    :type nodeid: int
    :type fmin: float
    :type fmax: float

    """

    _valid_steps = [
        "static",
        "eigenfrequency",
        "response_analysis",
        "dynamic",
        "complex_eig",
        "explicit",
    ]
    _valid_dyn_type = ["QUASI-STATIC", "TRANSIENT FIDELITY"]

    default_hist = HistOutput("default_hist", None, "energy", HistOutput.default_hist)
    default_field = FieldOutput("default_fields", int_type="FREQUENCY", int_value=1)

    def __init__(
        self,
        name,
        step_type,
        total_time=None,
        nl_geom=False,
        total_incr=1000,
        init_incr=100.0,
        min_incr=1e-8,
        max_incr=100.0,
        unsymm=False,
        stabilize=None,
        dyn_type="QUASI-STATIC",
        init_accel_calc=True,
        eigenmodes=20,
        alpha=0.1,
        beta=10,
        nodeid=None,
        fmin=0,
        restart_int=None,
        fmax=10,
        visco=None,
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        if step_type not in self._valid_steps:
            raise ValueError(f'Step type "{step_type}" is currently not supported')
        if total_time is not None:
            if init_incr > total_time and step_type != "explicit" and nl_geom is True:
                raise ValueError(f"Initial increment ({init_incr}) must be smaller than total time ({total_time})")
        else:
            total_time = init_incr
        if dyn_type not in Step._valid_dyn_type:
            raise ValueError(f'Dynamic input type "{dyn_type}" is not supported')

        self._restart_int = restart_int
        self._step_type = step_type
        self._nl_geom = nl_geom
        self._total_incr = total_incr
        self._init_incr = init_incr
        self._total_time = total_time
        self._min_incr = min_incr
        self._max_incr = max_incr
        self._unsymm = unsymm
        self._stabilize = stabilize
        self._dyn_type = dyn_type
        self._init_accel_calc = init_accel_calc
        self._eigenmodes = eigenmodes
        self._alpha = alpha
        self._beta = beta
        self._nodeid = nodeid
        self._fmin = fmin
        self._fmax = fmax
        self._visco = visco

        # Not-initialized parameters
        self._bcs = dict()
        self._loads = []
        self._interactions = dict()
        self._hist_outputs = [self.default_hist]
        self._field_outputs = [self.default_field]

    def add_load(self, load):
        """

        :param load:
        :type load: Load
        """
        self._loads.append(load)

    def add_bc(self, bc):
        """
        Adds a BC move

        :param bc: Boundary condition object
        :type bc: ada.fem.Bc

        """
        bc.parent = self
        self._bcs[bc.name] = bc

        if bc.fem_set not in self.parent.sets and bc.fem_set.parent is None:
            self.parent.sets.add(bc.fem_set)

    def add_history_output(self, hist_output):
        """
        Adds history output requests

        :param hist_output: Unique History Output
        :type hist_output: HistOutput
        """
        self._hist_outputs.append(hist_output)

    def add_field_output(self, field_output):
        """
        Adds field output requests

        :param field_output: Unique field output
        :type field_output: FieldOutput

        """
        self._field_outputs.append(field_output)

    def add_interaction(self, interaction):
        """

        :param interaction:
        :type interaction: Interaction
        :return:
        """
        interaction.parent = self
        self._interactions[interaction.name] = interaction

    @property
    def type(self):
        return self._step_type

    @property
    def nl_geom(self):
        return self._nl_geom

    @property
    def total_incr(self):
        return self._total_incr

    @property
    def init_incr(self):
        return self._init_incr

    @property
    def total_time(self):
        return self._total_time

    @property
    def min_incr(self):
        return self._min_incr

    @property
    def max_incr(self):
        return self._max_incr

    @property
    def unsymm(self):
        return self._unsymm

    @property
    def stabilize(self):
        """

        :return:
        """
        return self._stabilize

    @property
    def dyn_type(self):
        return self._dyn_type

    @property
    def init_accel_calc(self):
        return self._init_accel_calc

    @property
    def eigenmodes(self):
        return self._eigenmodes

    @property
    def alpha(self):
        return self._alpha

    @property
    def beta(self):
        return self._beta

    @property
    def nodeid(self):
        return self._nodeid

    @property
    def fmin(self):
        return self._fmin

    @property
    def fmax(self):
        return self._fmax

    @property
    def visco(self):
        return self._visco

    @property
    def interactions(self):
        return self._interactions

    @property
    def bcs(self):
        return self._bcs

    @property
    def restart_int(self):
        """

        :return: Restart request intervals
        """
        return self._restart_int

    @property
    def loads(self):
        return self._loads

    @property
    def field_outputs(self):
        return self._field_outputs

    @property
    def hist_outputs(self):
        return self._hist_outputs

    def __repr__(self):
        return f"Step({self.name}, type={self.type}, nl_geom={self.nl_geom})"


class LoadCase(FemBase):
    """
    The base LoadCase object. Defaults lc ids refer to the dictionary of basic loadcases used in ULS analysis.
    The dict can be changed using add_basic_lc function, but must be same as Genie lc FEMid.

    :param name:
    :param comment:
    :param loads:
    :param mass:
    :param lcsys:
    :param metadata:
    :param parent:
    """

    def __init__(
        self,
        name,
        comment,
        loads=None,
        mass=None,
        lcsys=None,
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        self._comment = comment
        self._loads = loads
        self._mass = mass
        self._lcsys = lcsys

    @property
    def loads(self):
        return self._loads

    @property
    def mass(self):
        return self._mass

    @property
    def comment(self):
        return self._comment

    @property
    def csys(self):
        return (1, 0, 0), (0, 1, 0), (0, 0, 1) if self._lcsys is None else self._lcsys

    def __repr__(self):
        return f"LC({self.name}, {self.comment})"


class Load(FemBase):
    TYPES = ["gravity", "acc", "acc_rot", "force", "force_set", "mass"]

    """


    :param load_type: Type of loads. See Load.TYPES for allowable load types.
    :param magnitude: Magnitude of load
    :param name: (Required in the event of a point load being applied)
    :param fem_set: Set reference (Required in the event of a point load being applied)
    :param dof: Degrees of freedom (Required in the event of a point load being applied)
    :param follower_force: Should follower force be accounted for
    :param amplitude: Attach an amplitude object to the load
    :param accr_origin: Origin of a rotational Acceleration field (necessary for load_type='acc_rot').
    :type load_type: str
    :type magnitude: float
    :type dof: list
    :type fem_set: FemSet
    """

    def __init__(
        self,
        name,
        load_type,
        magnitude,
        fem_set=None,
        dof=None,
        amplitude=None,
        follower_force=False,
        acc_vector=None,
        accr_origin=None,
        accr_rot_axis=None,
        csys=None,
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        self.type = load_type
        self._magnitude = magnitude
        self._fem_set = fem_set
        self._dof = dof
        self._amplitude = amplitude
        self._follower_force = follower_force
        self._acc_vector = acc_vector
        self._accr_origin = accr_origin
        self._accr_rot_axis = accr_rot_axis
        self._csys = csys
        if self.type == "point_load":
            if self._dof is None or self._fem_set is None or self._name is None:
                raise Exception("self._dofs and nid (Node id) and name needs to be set in order to use point loads")
            if len(self._dof) != 6:
                raise Exception(
                    "You need to include all 6 dofs even though forces are not applied in all 6 dofs. "
                    "Use None or 0.0 for the dofs not applied with forces"
                )

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        if value.lower() not in self.TYPES:
            raise ValueError(f'Load type "{value}" is not yet supported or does not exist. Must be "{self.TYPES}"')
        self._type = value

    @property
    def dof(self):
        if self._dof is None and self.type is co.GRAVITY:
            self._dof = [0, 0, 1] if self._dof is None else self._dof
        return self._dof

    @property
    def forces(self):
        if self.type not in ("force",):
            return None

        return [x * self.magnitude for x in self.dof]

    @property
    def forces_global(self):
        if self.type not in ("force",):
            return None
        if self.csys is None:
            return [x * self.magnitude for x in self.dof]
        else:
            csys = self.csys
            if csys.coords is None:
                logging.error("Calculating global forces without COORDS is not yet supported")
                return None

            from ada.core.utils import rotation_matrix_csys_rotate

            destination_csys = [(1, 0, 0), (0, 1, 0)]
            rmat = rotation_matrix_csys_rotate(csys.coords, destination_csys)
            res = np.concatenate([np.dot(rmat, np.array(self.dof[:3])), np.dot(rmat, np.array(self.dof[3:]))])
            return [x * self.magnitude for x in res]

    @property
    def amplitude(self):
        return self._amplitude

    @property
    def magnitude(self):
        return self._magnitude

    @property
    def fem_set(self):
        return self._fem_set

    @property
    def follower_force(self):
        return self._follower_force

    @property
    def acc_vector(self):
        if self.type not in ("acc", "acc_rot"):
            raise ValueError('Acceleration vector only applies for type "acc"')

        dir_error = "If acc_vector is not specified, you must pass dof=[int] (int 1-3) for the acc field"

        if self._acc_vector is not None:
            return self._acc_vector
        else:
            if len(self._dof) != 1:
                raise ValueError(dir_error)
            acc_dir = self._dof[0]
            if 1 > acc_dir > 3:
                raise ValueError(dir_error)

            if acc_dir == 1:
                dvec = 1, 0, 0
            elif acc_dir == 2:
                dvec = 0, 1, 0
            else:
                dvec = 0, 0, 1

            return tuple([float(self._magnitude * d) if d != 0 else 0.0 for d in dvec])

    @property
    def acc_rot_origin(self):
        return self._accr_origin

    @acc_rot_origin.setter
    def acc_rot_origin(self, value):
        self._accr_origin = value

    @property
    def acc_rot_axis(self):
        return self._accr_rot_axis

    @property
    def csys(self):
        """

        :rtype: Csys
        """
        return self._csys

    def __repr__(self):
        forc_str = ",".join(f"{f:.6E}" for f in self.forces)
        return f"Load({self.name}, {self.type}, [{forc_str}])"



import pathlib

import numpy as np
from IPython.display import display
from ipywidgets import Dropdown, HBox, VBox

from .common import (
    get_edges_and_faces_from_mesh,
    get_edges_from_fem,
    get_faces_from_fem,
    get_vertices_from_fem,
    magnitude,
)
from .renderer import MyRenderer
from .threejs_geom import edges_to_mesh, faces_to_mesh, vertices_to_mesh


def render_mesh(vertices, faces, colors):
    """
    Renders

    :param vertices:
    :param faces:
    :param colors:
    :return:
    """

    mesh = faces_to_mesh("faces", vertices, faces, colors)

    renderer = MyRenderer()
    renderer._displayed_pickable_objects.add(mesh)
    renderer.build_display()
    display(HBox([VBox([HBox(renderer._controls), renderer._renderer]), renderer.html]))


def viz_fem(fem, mesh, data_type):
    """

    :param fem:
    :type fem: ada.fem.FEM
    :param mesh:
    :type mesh: meshio.
    :param data_type:
    :type data_type:
    :return:
    :rtype:
    """
    u = np.asarray(mesh.point_data[data_type], dtype="float32")
    vertices = np.asarray(mesh.points, dtype="float32")
    faces = np.asarray(get_faces_from_fem(fem), dtype="uint16").ravel()

    res = [magnitude(u_) for u_ in u]
    max_r = max(res)
    res_norm_col = np.asarray([(x / max_r, 0, 0) for x in res], dtype="float32")

    render_mesh(vertices, faces, res_norm_col)


class Results:
    def __init__(self, result_file_path, part=None, palette=None):
        self.palette = [(0, 149 / 255, 239 / 255), (1, 0, 0)] if palette is None else palette
        self._analysis_type = None
        self._point_data = []
        self._cell_data = []
        self._results_file_path = result_file_path
        self._read_result_file(result_file_path)
        self._part = part
        self._renderer = None
        self._render_sets = None
        self._undeformed_mesh = None
        self._deformed_mesh = None

    @property
    def mesh(self):
        return self._mesh

    @property
    def renderer(self):
        """

        :return:
        :rtype: ada.base.renderer.MyRenderer
        """
        return self._renderer

    @property
    def mesh_undeformed(self):
        return self._undeformed_mesh

    @property
    def mesh_deformed(self):
        """

        :return:
        :rtype: pythreejs.Mesh
        """
        return self._deformed_mesh

    def _get_mesh(self, file_ref):
        import meshio

        file_ref = pathlib.Path(file_ref)
        if file_ref.suffix.lower() == ".rmed":
            mesh = meshio.read(file_ref, "med")
            self._analysis_type = "code_aster"
        else:
            mesh = meshio.read(file_ref)

        return mesh

    def _read_part(self, file_ref):
        mesh = self._get_mesh(file_ref)
        self._mesh = mesh
        self._vertices = np.asarray(get_vertices_from_fem(self._part.fem), dtype="float32")
        self._faces = np.asarray(get_faces_from_fem(self._part.fem), dtype="uint16").ravel()
        self._edges = np.asarray(get_edges_from_fem(self._part.fem), dtype="uint16").ravel()

        for n in mesh.point_data.keys():
            self._point_data.append(n)

        for n in mesh.cell_data.keys():
            self._cell_data.append(n)

    def _read_result_file(self, file_ref):
        mesh = self._get_mesh(file_ref)
        self._mesh = mesh
        self._vertices = np.asarray(mesh.points, dtype="float32")

        edges, faces = get_edges_and_faces_from_mesh(mesh)
        self._edges = np.asarray(edges, dtype="uint16").ravel()
        self._faces = np.asarray(faces, dtype="uint16").ravel()

        for n in mesh.point_data.keys():
            self._point_data.append(n)

        for n in mesh.cell_data.keys():
            self._cell_data.append(n)

    def _colorize_data(self, data, func=magnitude):
        res = [func(d) for d in data]
        sorte = sorted(res)
        min_r = sorte[0]
        max_r = sorte[-1]

        start = np.array(self.palette[0])
        end = np.array(self.palette[-1])

        def curr_p(t):
            return start + (end - start) * t / (max_r - min_r)

        colors = np.asarray([curr_p(x) for x in res], dtype="float32")
        return colors

    def create_viz_geom(self, data_type, displ_data=False, renderer=None):
        """

        :param data_type:
        :param displ_data:
        :type renderer: ada.base.renderer.MyRenderer
        :return:
        """
        default_vertex_color = (8, 8, 8)

        data = np.asarray(self.mesh.point_data[data_type], dtype="float32")
        colors = self._colorize_data(data)

        if renderer is None:
            renderer = MyRenderer()
            self._renderer = renderer

        # deformations
        if displ_data is True:
            vertices = np.asarray([x + u[:3] for x, u in zip(self._vertices, data)], dtype="float32")
            if self._undeformed_mesh is None:
                dark_grey = (0.66, 0.66, 0.66)
                white_color = np.asarray([dark_grey for x in self._vertices], dtype="float32")
                o_mesh = faces_to_mesh("undeformed", self._vertices, self._faces, white_color, opacity=0.5)
                self._undeformed_mesh = o_mesh
                renderer._displayed_non_pickable_objects.add(o_mesh)
        else:
            vertices = self._vertices
            if self._undeformed_mesh is not None:
                renderer._displayed_non_pickable_objects.remove(self._undeformed_mesh)
                self._undeformed_mesh = None

        vertices = np.array(vertices, dtype=np.float32)

        # Colours
        mesh = faces_to_mesh("deformed", vertices, self._faces, colors)
        points = vertices_to_mesh("deformed_vertices", vertices, default_vertex_color)
        lines = edges_to_mesh("deformed_lines", vertices, self._edges, default_vertex_color)

        if self._deformed_mesh is None:
            self._deformed_mesh = (mesh, points, lines)
            renderer._displayed_pickable_objects.add(mesh)
            renderer._displayed_pickable_objects.add(points)
            renderer._displayed_pickable_objects.add(lines)
            renderer.build_display(camera_type="perspective")
        else:
            face_geom = self._deformed_mesh[0].geometry
            face_geom.attributes["position"].array = vertices
            face_geom.attributes["index"].array = self._faces
            face_geom.attributes["color"].array = colors

            point_geom = self._deformed_mesh[1].geometry
            point_geom.attributes["position"].array = vertices

            edge_geom = self._deformed_mesh[2].geometry
            edge_geom.attributes["position"].array = vertices
            edge_geom.attributes["index"].array = self._edges

    def on_changed_point_data_set(self, p):
        data = p["new"]
        if self._analysis_type == "code_aster":
            if "point_tags" in data:
                print("\r" + "Point Tags are not a valid display value" + 10 * " ", end="")
                return None
            if "DISP" in data:
                self.create_viz_geom(data, displ_data=True, renderer=self.renderer)
            else:
                self.create_viz_geom(data, renderer=self.renderer)

    def _repr_html_(self):
        if self._renderer is None:
            self._renderer = MyRenderer()
            if self._analysis_type == "code_aster":
                data = [x for x in self._point_data if "DISP" in x][-1]
                self.create_viz_geom(data, displ_data=True, renderer=self.renderer)
                i = self._point_data.index(data)
                self._render_sets = Dropdown(
                    options=self._point_data, value=self._point_data[i], tooltip="Select a set", disabled=False
                )
                self._render_sets.observe(self.on_changed_point_data_set, "value")
                self.renderer._controls.pop()
                self.renderer._controls.append(self._render_sets)
            else:
                raise NotImplementedError(f'Support for analysis_type "{self._analysis_type}"')

        display(HBox([VBox([HBox(self.renderer._controls), self.renderer._renderer]), self.renderer.html]))

    def __repr__(self):
        return f"Results({self._analysis_type}, {self._results_file_path.name})"



import numpy as np


def format_color(r, g, b):
    return "#%02x%02x%02x" % (r, g, b)


def get_edges_and_faces_from_mesh(mesh):
    """

    :param mesh:
    :type mesh: meshio.Mesh
    :return:
    """
    from ada.fem import ElemShapes
    from ada.fem.io.io_meshio import meshio_to_ada_type

    edges = []
    faces = []
    for cell_block in mesh.cells:
        el_type = meshio_to_ada_type[cell_block.type]
        for elem in cell_block.data:
            res = ElemShapes(el_type, elem)
            edges += res.edges
            faces += res.faces
    return edges, faces


def get_faces_from_fem(fem):
    """

    :param fem:
    :type fem: ada.fem.FEM
    :return:
    :rtype: list
    """
    ids = []
    for el in fem.elements.elements:
        for f in el.shape.faces:
            # Convert to indices, not id
            ids += [[int(e.id - 1) for e in f]]
    return ids


def get_edges_from_fem(fem):
    """

    :param fem:
    :type fem: ada.fem.FEM
    :return:
    :rtype: list
    """
    ids = []
    for el in fem.elements.elements:
        for f in el.shape.edges_seq:
            # Convert to indices, not id
            ids += [[int(el.nodes[e].id - 1) for e in f]]
    return ids


def get_vertices_from_fem(fem):
    """

    :param fem:
    :type fem: ada.fem.FEM
    :return:
    """

    return np.asarray([n.p for n in fem.nodes._nodes], dtype="float32")


def get_bounding_box(vertices):
    return np.min(vertices, 0), np.max(vertices, 0)


def magnitude(u):
    return np.sqrt(u[0] ** 2 + u[1] ** 2 + u[2] ** 2)



import os
import pathlib
from itertools import groupby
from operator import attrgetter

import numpy as np


def write_assembly_to_cache(assembly, cache_file_path):
    """
    Write the Assembly information to a HDF5 file format for High performance cache.

    TODO: Add support for FEM only to begin with

    :param assembly:
    :param cache_file_path:
    :type assembly: ada.Assembly
    :return:
    """
    import h5py

    cache_file_path = pathlib.Path(cache_file_path)
    os.makedirs(cache_file_path.parent, exist_ok=True)
    h5_filename = cache_file_path.with_suffix(".h5")
    f = h5py.File(h5_filename, "w")

    info = f.create_group("INFO")
    info.attrs.create("NAME", assembly.name)
    parts_group = f.create_group("PARTS")
    for p in assembly.get_all_parts_in_assembly(True):
        add_part_to_cache(p, parts_group)

    f.close()

    print(f'Saved cached model at "{cache_file_path}"')


def add_part_to_cache(part, parts_group):
    """

    :param part:
    :type part: ada.Part
    :param parts_group:
    """
    part_group = parts_group.create_group(part.name)
    # Add Materials, Sections
    # Add Beams, Plates, etc..

    # Add FEM object
    if len(part.fem.nodes) > 0:
        add_fem_to_cache(part.fem, part_group)


def add_fem_to_cache(fem, part_group):
    """

    :param fem:
    :type fem: ada.fem.FEM
    :param part_group:
    """
    fem_group = part_group.create_group("FEM")
    fem_group.attrs.create("NAME", fem.name)

    # Add Nodes
    points = np.zeros((int(fem.nodes.max_nid), 3))

    def pmap(n):
        points[int(n.id - 1)] = n.p

    list(map(pmap, fem.nodes))

    coo = fem_group.create_dataset("NODES", data=points)
    coo.attrs.create("NBR", len(points))
    coo.attrs.create("IDS", [p.id for p in fem.nodes])

    # Add elements
    def get_node_ids_from_element(el_):
        return [int(n.id) for n in el_.nodes]

    elements_group = fem_group.create_group("MESH")

    for group, elements in groupby(fem.elements, key=attrgetter("type")):
        elements = list(elements)
        med_cells = elements_group.create_group(group)
        elements = [[int(el.id), *get_node_ids_from_element(el)] for el in elements]
        med_cells.create_dataset("ELEMENTS", data=elements)



from __future__ import absolute_import, division, print_function

import time
import uuid

from ifcopenshell import main
from ifcopenshell.file import file as ifc_file
from ifcopenshell.guid import compress

# A quick way to setup an 'empty' IFC file, taken from:
# http://academy.ifcopenshell.org/creating-a-simple-wall-with-property-set-and-quantity-information/
TEMPLATE = """ISO-10303-21;
HEADER;
FILE_DESCRIPTION(('ViewDefinition [CoordinationView]'),'2;1');
FILE_NAME('%(filename)s','%(timestring)s',('%(creator)s'),('%(organization)s'),'%(application)s','%(application)s','');
FILE_SCHEMA(('%(schema_identifier)s'));
ENDSEC;
DATA;
#1=IFCPERSON($,$,'%(creator)s',$,$,$,$,$);
#2=IFCORGANIZATION($,'%(organization)s',$,$,$);
#3=IFCPERSONANDORGANIZATION(#1,#2,$);
#4=IFCAPPLICATION(#2,'%(application_version)s','%(application)s','');
#5=IFCOWNERHISTORY(#3,#4,$,.ADDED.,$,#3,#4,%(timestamp)s);
#6=IFCDIRECTION((1.,0.,0.));
#7=IFCDIRECTION((0.,0.,1.));
#8=IFCCARTESIANPOINT((0.,0.,0.));
#9=IFCAXIS2PLACEMENT3D(#8,#7,#6);
#10=IFCDIRECTION((0.,1.,0.));
#11=IFCGEOMETRICREPRESENTATIONCONTEXT($,'Model',3,1.E-05,#9,#10);
#12=IFCDIMENSIONALEXPONENTS(0,0,0,0,0,0,0);
#13=IFCSIUNIT(*,.LENGTHUNIT.,%(units_str)s);
#14=IFCSIUNIT(*,.AREAUNIT.,$,.SQUARE_METRE.);
#15=IFCSIUNIT(*,.VOLUMEUNIT.,$,.CUBIC_METRE.);
#16=IFCSIUNIT(*,.PLANEANGLEUNIT.,$,.RADIAN.);
#17=IFCMEASUREWITHUNIT(IFCPLANEANGLEMEASURE(0.017453292519943295),#16);
#18=IFCCONVERSIONBASEDUNIT(#12,.PLANEANGLEUNIT.,'DEGREE',#17);
#19=IFCUNITASSIGNMENT((#13,#14,#15,#18));
#20=IFCPROJECT('%(project_globalid)s',#5,'%(project_name)s',$,$,$,$,(#11),#19);
ENDSEC;
END-ISO-10303-21;
"""

DEFAULTS = {
    "application": lambda d: "IfcOpenShell-%s" % main.version,
    "application_version": lambda d: main.version,
    "project_globalid": lambda d: compress(uuid.uuid4().hex),
    "schema_identifier": lambda d: main.schema_identifier,
    "timestamp": lambda d: int(time.time()),
    "timestring": lambda d: time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime(d.get("timestamp") or time.time())),
}


def tpl_create(
    filename=None,
    timestring=None,
    organization=None,
    creator=None,
    schema_identifier=None,
    application_version=None,
    timestamp=None,
    application=None,
    project_globalid=None,
    project_name=None,
    units_str=None,
):
    d = dict(locals())

    def _():
        for var, value in d.items():
            if value is None:
                yield var, DEFAULTS.get(var, lambda *args: "")(d)

    d.update(dict(_()))

    return ifc_file.from_string(TEMPLATE % d)



import os
import pathlib


def _get_platform_home():
    """

    :return:
    """
    # _platform_home = dict(win32="C:/ADA", linux="/home/ADA", linux2="/home/ADA", macos="/home/ADA")
    # return _platform_home[sys.platform]

    return pathlib.Path.home() / "ADA"


class Settings:
    """
    The Properties object contains all general purpose properties relevant for Parts and Assemblies

    """

    point_tol = 1e-4
    precision = 6
    mtol = 1e-3
    mmtol = 1
    valid_units = ["m", "mm"]

    convert_bad_names = False
    use_occ_bounding_box_algo = False
    use_oriented_bbox = True  # This is only relevant if OCC bounding box is True
    make_param_elbows = False

    # IFC export settings
    include_ecc = True

    # FEM analysis settings
    if os.getenv("ADA_execute_dir", None) is not None:
        execute_dir = pathlib.Path(os.getenv("ADA_execute_dir", None))
    else:
        execute_dir = None

    # Code Aster conversion specific settings
    ca_use_meshio_med_convert = False
    ca_experimental_id_numbering = False

    # Fem Results
    return_experimental_fem_res_after_execute = False

    # Execution Settings
    use_docker_execute = False

    debug = False
    _home = _get_platform_home()
    scratch_dir = pathlib.Path(os.getenv("ADA_scratch_dir", f"{_home}/Analyses"))
    temp_dir = pathlib.Path(os.getenv("ADA_temp_dir", f"{_home}/temp"))
    debug_dir = pathlib.Path(os.getenv("ADA_log_dir", f"{_home}/logs"))
    test_dir = pathlib.Path(os.getenv("ADA_test_dir", f"{_home}/tests"))
    tools_dir = pathlib.Path(os.getenv("ADA_tools_dir", f"{_home}/tools"))

    fem_exe_paths = dict(abaqus=None, ccx=None, sesam=None, usfos=None, code_aster=None)

    @classmethod
    def default_ifc_settings(cls):
        import ifcopenshell.geom

        ifc_settings = ifcopenshell.geom.settings()
        ifc_settings.set(ifc_settings.USE_PYTHON_OPENCASCADE, True)
        ifc_settings.set(ifc_settings.SEW_SHELLS, True)
        ifc_settings.set(ifc_settings.WELD_VERTICES, True)
        ifc_settings.set(ifc_settings.INCLUDE_CURVES, True)
        ifc_settings.set(ifc_settings.USE_WORLD_COORDS, True)
        ifc_settings.set(ifc_settings.VALIDATE_QUANTITIES, True)
        return ifc_settings



import logging
import os
import pathlib

from ada.config import Settings as _Settings
from ada.core.constants import color_map as _cmap


class Backend:
    def __init__(self, name, guid=None, metadata=None, units="m", parent=None, ifc_settings=None):
        self.name = name
        self._parent = parent
        self._ifc_settings = ifc_settings
        from ada.core.utils import create_guid

        self.guid = create_guid() if guid is None else guid
        units = units.lower()
        if units not in _Settings.valid_units:
            raise ValueError(f'Unit type "{units}"')
        self._units = units
        self._metadata = metadata if metadata is not None else dict()

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if _Settings.convert_bad_names:
            logging.debug("Converting bad name")
            value = value.replace("/", "_").replace("=", "")
            if str.isnumeric(value[0]):
                value = "ADA_" + value

        if "/" in value:
            logging.debug(f'Character "/" found in {value}')

        self._name = value.strip()

    @property
    def guid(self):
        return self._guid

    @guid.setter
    def guid(self, value):
        if value is None:
            raise ValueError("guid cannot be None")
        self._guid = value

    @property
    def parent(self):
        """

        :return:
        :rtype: ada.Part
        """
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def metadata(self):
        return self._metadata

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        raise NotImplementedError("Assigning units is not yet represented for this object")

    @staticmethod
    def _unit_conversion(ori_unit, value):
        if value == "m" and ori_unit == "mm":
            scale_factor = 0.001
        elif value == "mm" and ori_unit == "m":
            scale_factor = 1000
        else:
            raise ValueError(f'Unrecognized unit conversion from "{ori_unit}" to "{value}"')
        return scale_factor

    @property
    def ifc_settings(self):
        if self._ifc_settings is None:
            import ifcopenshell.geom

            ifc_settings = ifcopenshell.geom.settings()
            ifc_settings.set(ifc_settings.USE_PYTHON_OPENCASCADE, True)
            ifc_settings.set(ifc_settings.SEW_SHELLS, True)
            ifc_settings.set(ifc_settings.WELD_VERTICES, True)
            ifc_settings.set(ifc_settings.INCLUDE_CURVES, True)
            ifc_settings.set(ifc_settings.USE_WORLD_COORDS, True)
            ifc_settings.set(ifc_settings.VALIDATE_QUANTITIES, True)
            self._ifc_settings = ifc_settings
        return self._ifc_settings

    @ifc_settings.setter
    def ifc_settings(self, value):
        self._ifc_settings = value

    def get_assembly(self):
        """

        :return:
        :rtype: Assembly
        """
        from ada import Assembly

        parent = self
        still_looking = True
        max_levels = 10
        a = 0
        while still_looking is True:
            if issubclass(type(parent), Assembly) is True:
                still_looking = False
            if parent.parent is None:
                break
            a += 1
            parent = parent.parent
            if a > max_levels:
                break
        return parent


class BackendGeom(Backend):
    """
    The shared backend of all physical components (Beam, Plate) or aggregate of components (Part, Assembly).

    """

    _renderer = None

    def __init__(self, name, guid=None, metadata=None, units="m", parent=None, colour=None):
        super().__init__(name, guid, metadata, units, parent)
        self._penetrations = []
        self.colour = colour

    def add_penetration(self, pen):
        """

        :param pen:
        """
        from ada import Penetration, PrimBox, PrimCyl, PrimExtrude, PrimRevolve

        if type(pen) not in [Penetration, PrimExtrude, PrimRevolve, PrimCyl, PrimBox]:
            raise ValueError(f'Unsupported penetration type "{type(pen)}"')

        pen.parent = self
        if type(pen) in (PrimExtrude, PrimRevolve, PrimCyl, PrimBox):
            self._penetrations.append(Penetration(pen, parent=self))
        else:
            self._penetrations.append(pen)

        return pen

    def to_stp(self, destination_file, geom_repr="solid", schema="AP242"):
        """
        Write current assembly to STEP file

        OpenCascade reference:

            https://www.opencascade.com/doc/occt-7.4.0/overview/html/occt_user_guides__step.html#occt_step_3


        :param destination_file:
        :param geom_repr:
        :param schema: STEP Schemas.
        """

        from OCC.Core.IFSelect import IFSelect_RetError
        from OCC.Core.Interface import Interface_Static_SetCVal
        from OCC.Core.STEPConstruct import stepconstruct_FindEntity
        from OCC.Core.STEPControl import STEPControl_AsIs, STEPControl_Writer
        from OCC.Core.TCollection import TCollection_HAsciiString

        from ada.core.utils import Counter

        if geom_repr not in ["shell", "solid"]:
            raise ValueError('Geometry representation can only accept either "solid" or "shell" as input')

        destination_file = pathlib.Path(destination_file).with_suffix(".stp")

        assembly_mode = 1
        shp_names = Counter(1, "shp")
        writer = STEPControl_Writer()
        fp = writer.WS().TransferWriter().FinderProcess()
        Interface_Static_SetCVal("write.step.schema", schema)
        # Interface_Static_SetCVal('write.precision.val', '1e-5')
        Interface_Static_SetCVal("write.precision.mode", "1")
        Interface_Static_SetCVal("write.step.assembly", str(assembly_mode))

        from ada import Assembly, Beam, Part, Pipe, Plate, Shape, Wall

        def add_geom(geo, o):
            name = o.name if o.name is not None else next(shp_names)
            Interface_Static_SetCVal("write.step.product.name", name)
            stat = writer.Transfer(geo, STEPControl_AsIs)
            if int(stat) > int(IFSelect_RetError):
                raise Exception("Some Error occurred")

            item = stepconstruct_FindEntity(fp, geo)
            if not item:
                logging.debug("STEP item not found for FindEntity")
            else:
                item.SetName(TCollection_HAsciiString(name))

        if type(self) is Shape:
            assert isinstance(self, Shape)
            add_geom(self.geom, self)
        elif type(self) in (Beam, Plate, Wall):
            assert isinstance(self, (Beam, Plate, Wall))
            if geom_repr == "shell":
                add_geom(self.shell, self)
            else:
                add_geom(self.solid, self)
        elif type(self) is Pipe:
            assert isinstance(self, Pipe)
            for geom in self.geometries:
                add_geom(geom, self)
        elif type(self) in (Part, Assembly):
            assert isinstance(self, Part)

            for p in self.get_all_subparts() + [self]:
                for obj in list(p.plates) + list(p.beams) + list(p.shapes) + list(p.pipes) + list(p.walls):
                    if type(obj) in (Plate, Beam, Wall):
                        try:
                            if geom_repr == "shell":
                                add_geom(obj.shell, self)
                            else:
                                add_geom(obj.solid, self)
                        except BaseException as e:
                            logging.info(f'passing pl "{obj.name}" due to {e}')
                            continue
                    elif type(obj) in (Pipe,):
                        assert isinstance(obj, Pipe)
                        for geom in obj.geometries:
                            add_geom(geom, self)
                    elif type(obj) is Shape:
                        add_geom(obj.geom, self)
                    else:
                        raise ValueError("Unkown Geometry type")

        os.makedirs(destination_file.parent, exist_ok=True)

        status = writer.Write(str(destination_file))
        if int(status) > int(IFSelect_RetError):
            raise Exception("Error during write operation")

        print(f'step file created at "{destination_file}"')

    def render_locally(
        self, addr="localhost", server_port=8080, open_webbrowser=False, render_engine="threejs", resolution=(1800, 900)
    ):
        from OCC.Display.WebGl.simple_server import start_server

        if render_engine == "xdom":
            from OCC.Display.WebGl import x3dom_renderer

            my_renderer = x3dom_renderer.X3DomRenderer()
            # TODO: Find similarities in build processing as done for THREE.js (tesselate geom etc..).
            # my_renderer.DisplayShape(shape.profile_curve_outer.wire)
            # my_renderer.DisplayShape(shape.sweep_curve.wire)
            # my_renderer.DisplayShape(shape.geom)
            my_renderer.render()
        else:  # Assume THREEJS
            from ipywidgets.embed import embed_minimal_html

            from ada.base.renderer import MyRenderer

            _path = pathlib.Path("temp/index.html").resolve().absolute()

            renderer = MyRenderer(resolution)
            renderer.DisplayObj(self)
            renderer.build_display()

            os.makedirs(_path.parent, exist_ok=True)
            embed_minimal_html(_path, views=renderer._renderer, title="Pythreejs Viewer")
            start_server(addr, server_port, str(_path.parent), open_webbrowser)

    def get_render_snippet(self, view_size=None):
        """
        Return the html snippet containing threejs renderer
        """
        from ipywidgets.embed import embed_snippet

        from ada.base.renderer import MyRenderer

        renderer = MyRenderer()
        renderer.DisplayObj(self)
        renderer.build_display()

        return embed_snippet(renderer._renderer)

    @property
    def colour(self):
        return self._colour

    @colour.setter
    def colour(self, value):
        if type(value) is str:
            if value.lower() not in _cmap.keys():
                raise ValueError("Currently unsupported")
            self._colour = _cmap[value.lower()]
        else:
            self._colour = value

    @property
    def colour_webgl(self):
        from OCC.Display.WebGl.jupyter_renderer import format_color

        if self.colour is None:
            return None
        if self.colour[0] == -1 and self.colour[1] == -1 and self.colour[2] == -1:
            return None

        if self.colour[0] <= 1.0:
            colour = [int(x * 255) for x in self.colour]
        else:
            colour = [int(x) for x in self.colour]

        colour_formatted = format_color(*colour)
        return colour_formatted

    @property
    def penetrations(self):
        return self._penetrations

    def _repr_html_(self):
        from IPython.display import display
        from ipywidgets import HBox, VBox

        from ada.base.renderer import MyRenderer

        renderer = MyRenderer()

        renderer.DisplayObj(self)
        renderer.build_display()
        self._renderer = renderer
        display(HBox([VBox([HBox(renderer._controls), renderer._renderer]), renderer.html]))
        # renderer._reset()
        # self._renderer.Display()
        return ""



import json
import logging
import os
import pathlib
import shutil
import subprocess

from ada.config import Settings as _Settings


def run_abaqus(
    inp_path,
    cpus=2,
    gpus=None,
    run_ext=False,
    manifest=None,
    subr_path=None,
    execute=True,
    return_bat_str=False,
    exit_on_complete=True,
):
    """

    :param inp_path: Path to input file folder(s)
    :param cpus: Number of CPUs to run the analysis on. Default is 2.
    :param gpus: Number of GPUs to run the analysis on. Default is none.
    :param run_ext: If False the process will wait for the abaqus analysis to finish. Default is False
    :param manifest: Dictionary containing various metadata relevant for the analysis
    :param subr_path: Path to fortran subroutine file (optional).
    :param execute: Automatically starts Abaqus analysis. Default is True
    :param return_bat_str:
    :param exit_on_complete:


    Note!
        'analysis_name' be name of input file if 'analysis_path' is pointing to containing folder.
        Alternatively if name of input file and containing folder shares the same name 'analysis_path' can be set to the
        top folder and analysis_name will refer to the subfolder and inp-file sharing the same name.

    """
    from ..utils import get_exe_path

    try:
        aba_dir = get_exe_path("abaqus")
    except FileNotFoundError as e:
        logging.error(e)
        return

    inp_path = pathlib.Path(inp_path)

    if inp_path.exists() is False:
        raise FileNotFoundError(f'Unable to find inp file "{inp_path}"')

    gpus = "" if gpus is None else f"GPUS={gpus}"
    analysis_name = inp_path.name.replace(".inp", "")

    if subr_path is None:
        param = ["job=" + analysis_name, "CPUS=" + str(cpus), gpus, "interactive"]
        call_str = f"call {aba_dir}"
    else:
        subr_path = pathlib.Path(subr_path)
        subr_name = subr_path.with_suffix("").name
        shutil.copy(subr_path, inp_path.parent / subr_path.name)
        subr = f"user={subr_name}"
        param = ["job=" + analysis_name, subr, "CPUS=" + str(cpus), gpus, "interactive"]
        prog = r"C:\Program Files (x86)"
        call_str = rf'''
call "{prog}\IntelSWTools\compilers_and_libraries_2018.3.210\windows\bin\ipsxe-comp-vars.bat" intel64 vs2017
call "{prog}\Microsoft Visual Studio 11.0\VC\bin\amd64\vcvars64.bat" intel 64
call "C:\SIMULIA\CAE\2017\win_b64\code\bin\ABQLauncher.exe"'''
    param_str = " ".join([str(val) for val in param if str(val) != ""])

    start_bat = "runABA.bat"
    stop_bat = "stopABA.bat"

    bat_start_str = f"""echo OFF
for %%* in (.) do set CurrDirName=%%~nx*
title %CurrDirName%
cd /d {inp_path.parent}
echo ON
{call_str} {param_str}"""

    if exit_on_complete:
        bat_start_str += "\nEXIT\nEXIT"

    with open(inp_path.parent / start_bat, "w") as d:
        d.write(bat_start_str)

    with open(inp_path.parent / stop_bat, "w") as d:
        d.write(f"cd /d {inp_path.parent}\nabaqus terminate job={analysis_name}")

    if manifest is not None:
        with open(inp_path.parent / "analysis_manifest.json", "w") as fp:
            json.dump(manifest, fp, indent=4)

    execute_path = _Settings.execute_dir if _Settings.execute_dir is not None else inp_path.parent

    if inp_path.parent != execute_path:
        os.makedirs(execute_path, exist_ok=True)
        shutil.copy(inp_path.parent / start_bat, execute_path / start_bat)
        shutil.copy(inp_path.parent / stop_bat, execute_path / stop_bat)

    if return_bat_str is True:
        return bat_start_str

    if execute is True:
        if run_ext is True:
            print(80 * "-")
            print(f'starting Abaqus simulation "{analysis_name}"')
            print(f"\nUsing the following parameters:\n{param_str}\n")
            subprocess.call("start " + start_bat, cwd=execute_path, shell=True)
            print("Note! This starts Abaqus in an external window on a separate thread.")
            print(80 * "-")
        else:
            print(80 * "-")
            print(f'starting Abaqus simulation "{analysis_name}"')
            print(f"\nUsing the following parameters:\n{param_str}\n")
            subprocess.call("start /wait " + start_bat, cwd=execute_path, shell=True)
            print(f'Finished Abaqus simulation "{analysis_name}"')
            print(80 * "-")



from ada import Beam, Part, Plate, Section
from ada.core.utils import Counter

bm_name = Counter(1, "bm")
pl_name = Counter(1, "pl")
floor_name = Counter(1, "floor")


class ReinforcedFloor(Part):
    def __init__(self, name, plate, spacing=0.2, s_type="HP140x8", stringer_dir="X"):
        """

        :param name:
        :param plate:
        :param spacing:
        :param s_type:
        :param stringer_dir:
        :type plate: ada.Plate
        """
        super(ReinforcedFloor, self).__init__(name)
        self.add_plate(plate)

        # Calculate number of stringers
        bbox = plate.bbox
        xmin, xmax = bbox[0]
        ymin, ymax = bbox[1]
        zmin, zmax = bbox[2]

        if stringer_dir == "Y":
            snum = int((xmax - xmin) / spacing) - 1
        else:
            snum = int((ymax - ymin) / spacing) - 1

        origin = plate.poly.origin
        z = origin[2]
        x = xmin + spacing
        y = ymin + spacing
        for i in range(0, snum):
            if stringer_dir == "Y":
                p1 = (x, ymin, z)
                p2 = (x, ymax, z)
                x += spacing
            else:
                p1 = (xmin, y, z)
                p2 = (xmax, y, z)
                y += spacing
            self.add_beam(Beam(next(bm_name), p1, p2, sec=s_type))

    def penetration_check(self):
        import numpy as np

        from ada import PipeSegStraight, PrimCyl

        a = self.get_assembly()
        cog = self.nodes.vol_cog
        normal = self._lz
        for p in a.get_all_subparts():
            for pipe in p.pipes:
                for segment in pipe.segments:
                    if type(segment) is PipeSegStraight:
                        assert isinstance(segment, PipeSegStraight)
                        p1, p2 = segment.p1, segment.p2
                        v1 = (p1.p - cog) * normal
                        v2 = (p2.p - cog) * normal
                        if np.dot(v1, v2) < 0:
                            self.add_penetration(
                                PrimCyl(f"{p.name}_{pipe.name}_{segment.name}_pen", p1.p, p2.p, pipe.section.r + 0.1)
                            )


class SimpleStru(Part):
    class Params:
        w = None
        l = None
        h = None

    def __init__(self, name, origin=(0, 0, 0), w=5, l=5, h=3, gsec="IPE200", csec="HEB200", pl_thick=10e-3):
        super(SimpleStru, self).__init__(name, origin)
        self.Params.w = w
        self.Params.h = h
        self.Params.l = l

        # Define the 5 corner points of each storey
        c1, c2, c3, c4 = self.c1, self.c2, self.c3, self.c4

        # Define the relationship of corners that make up the 4 support beams
        beams = [(c1, c2), (c2, c3), (c3, c4), (c4, c1)]

        z0 = origin[2]
        sec = Section(gsec, from_str=gsec, parent=self)
        for elev in [z0, h]:
            for p1, p2 in beams:
                self.add_beam(Beam(next(bm_name), n1=p1(elev), n2=p2(elev), sec=sec, jusl="TOP"))
            plate = Plate(next(pl_name), [c1(elev), c2(elev), c3(elev), c4(elev)], pl_thick, use3dnodes=True)
            self.add_part(ReinforcedFloor(next(floor_name), plate))

        # Columns
        columns = [(c1(z0), c1(h)), (c2(z0), c2(h)), (c3(z0), c3(h)), (c4(z0), c4(h))]
        for p1, p2 in columns:
            self.add_beam(Beam(next(bm_name), n1=p1, n2=p2, sec=csec))

    def c1(self, z):
        return 0, 0, z

    def c2(self, z):
        return self.Params.w, 0, z

    def c3(self, z):
        return self.Params.w, self.Params.l, z

    def c4(self, z):
        return 0, self.Params.l, z

    def add_bcs(self):
        from ada.fem import Bc, FemSet

        for i, bc_loc in enumerate([self.c1, self.c2, self.c3, self.c4]):
            fem_set_btn = FemSet(f"fix{i}", [], "nset")
            self.fem.add_set(fem_set_btn, p=bc_loc(self._origin[2]))
            self.fem.add_bc(Bc(f"bc_fix{i}", fem_set_btn, [1, 2, 3]))


def make_it_complex():
    from ada import Assembly, Pipe, Section

    a = Assembly("ParametricSite")

    pm = SimpleStru("ParametricModel")
    a.add_part(pm)

    elev = pm.Params.h - 0.4
    offset_y = 0.4
    pipe1 = Pipe(
        "Pipe1",
        [
            (0, offset_y, elev),
            (pm.Params.w + 0.4, offset_y, elev),
            (pm.Params.w + 0.4, pm.Params.l + 0.4, elev),
            (pm.Params.w + 0.4, pm.Params.l + 0.4, 0.4),
            (0, pm.Params.l + 0.4, 0.4),
        ],
        Section("PSec1", "PIPE", r=0.1, wt=10e-3),
    )

    pipe2 = Pipe(
        "Pipe2",
        [
            (0.5, offset_y + 0.5, elev + 1.4),
            (0.5, offset_y + 0.5, elev),
            (0.2 + pm.Params.w, offset_y + 0.5, elev),
            (0.2 + pm.Params.w, pm.Params.l + 0.4, elev),
            (0.2 + pm.Params.w, pm.Params.l + 0.4, 0.6),
            (0, pm.Params.l + 0.4, 0.6),
        ],
        Section("PSec2", "PIPE", r=0.05, wt=5e-3),
    )
    pm.add_part(Part("Piping") / [pipe1, pipe2])
    for p in pm.parts.values():
        if type(p) is ReinforcedFloor:
            p.penetration_check()

    return a



import numpy


def cell_data_from_raw(cells, cell_data_raw):
    cs = numpy.cumsum([len(c[1]) for c in cells])[:-1]
    return {name: numpy.split(d, cs) for name, d in cell_data_raw.items()}


def attribute_type(data):
    # <http://www.xdmf.org/index.php/XDMF_Model_and_Format#Attribute>
    if len(data.shape) == 1 or (len(data.shape) == 2 and data.shape[1] == 1):
        return "Scalar"
    elif len(data.shape) == 2 and data.shape[1] in [2, 3]:
        return "Vector"
    elif (len(data.shape) == 2 and data.shape[1] == 9) or (
        len(data.shape) == 3 and data.shape[1] == 3 and data.shape[2] == 3
    ):
        return "Tensor"
    elif len(data.shape) == 2 and data.shape[1] == 6:
        return "Tensor6"

    # if len(data.shape) != 3:
    #     raise ReadError()
    return "Matrix"


dtype_to_format_string = {
    "int32": "%d",
    "int64": "%d",
    "unit32": "%d",
    "uint64": "%d",
    "float32": "%.7e",
    "float64": "%.16e",
}


def raw_from_cell_data(cell_data):
    return {name: numpy.concatenate(value) for name, value in cell_data.items()}


class ReadError(Exception):
    pass


numpy_to_xdmf_dtype = {
    "int32": ("Int", "4"),
    "int64": ("Int", "8"),
    "uint32": ("UInt", "4"),
    "uint64": ("UInt", "8"),
    "float32": ("Float", "4"),
    "float64": ("Float", "8"),
}
xdmf_to_numpy_type = {v: k for k, v in numpy_to_xdmf_dtype.items()}



import functools
import logging
import os
from itertools import chain

import numpy as np

from ada import Node, Plate
from ada.core.containers import Nodes
from ada.core.utils import clockwise, intersect_calc, roundoff, vector_length
from ada.fem import Elem, FemSection, FemSet
from ada.fem.containers import FemElements

gmsh_map = {"Triangle 3": "S3", "Quadrilateral 4": "S4R"}


class GMesh:
    """
    A class wrapping around the gmsh python mesh library

    links to gmsh tips:

        https://forum.freecadweb.org/viewtopic.php?style=4&f=18&t=20796&start=30

    :param work_dir: Work directory
    """

    mesh_map = {"bm1", "B31", "bm2", "B32"}

    def __init__(self, part, work_dir="gmsh", tol=1e-5):
        self._part = part
        self._work_dir = os.path.abspath(work_dir)
        self._tol = tol
        self._bm_map = dict()
        self._pl_map = dict()

    def mesh(
        self,
        size=0.1,
        order=1,
        max_dim=2,
        interactive=False,
        mesh_algo=8,
        sh_int_points=5,
    ):
        """

        :param size:
        :param order:
        :param max_dim:
        :param interactive:
        :param mesh_algo:
        :param sh_int_points:
        :return:
        """
        import gmsh

        part = self._part
        pl_in = [pl_ for p in part.get_all_subparts() for pl_ in p.plates] + [pl for pl in part.plates]
        bm_in = [bm_ for p in part.get_all_subparts() for bm_ in p.beams] + [bm for bm in part.beams]
        try:
            gmsh.finalize()
        except BaseException as e:
            logging.debug(e)
        gmsh.initialize()
        gmsh.option.setNumber("General.Terminal", 1)
        gmsh.option.setNumber("Mesh.SecondOrderIncomplete", 1)
        gmsh.option.setNumber("Mesh.Algorithm", mesh_algo)
        gmsh.option.setNumber("Mesh.ElementOrder", order)

        list(
            map(
                functools.partial(
                    self._create_plate_geom,
                    beams=bm_in,
                    size=size,
                    interactive=interactive,
                ),
                pl_in,
            )
        )
        gmsh.model.geo.synchronize()

        list(
            map(
                functools.partial(self._create_bm_geom, size=size),
                filter(lambda x: x not in [b for b in self._bm_map.values()], bm_in),
            )
        )
        gmsh.model.geo.synchronize()

        if len(pl_in) > 0:
            gmsh.model.mesh.setRecombine(2, 1)

        gmsh.model.mesh.generate(max_dim)
        gmsh.model.mesh.removeDuplicateNodes()

        if interactive:
            gmsh.fltk.run()

        nodes = list(gmsh.model.mesh.getNodes(-1, -1))

        # Extract Gmsh model information and import the data into a FEM model
        self._part.fem._nodes = Nodes(list(map(self.get_mesh_nodes, nodes[0])), parent=self._part.fem)
        bm_elems = map(functools.partial(self.get_beam_elements, order=order), self._bm_map.keys())
        pl_elems = map(functools.partial(self.get_shell_elements, order=order), self._pl_map.keys())
        self._part.fem._elements = FemElements(
            chain.from_iterable(list(bm_elems) + list(pl_elems)), fem_obj=self._part.fem
        )
        self._part.fem.elements.renumber()

        gmsh.finalize()

    def mesh_shape(self, shp, size=0.1, order=1, max_dim=3, mesh_algo=8, interactive=False):
        """



        :param shp: ADA shape object
        :param size: Size of mesh
        :param order: Polynomial order
        :param max_dim: Maximum dimensions (1, 2 or 3).
        :param mesh_algo: 2D mesh algorithm (1: MeshAdapt, 2: Automatic, 5: Delaunay, 6: Frontal-Delaunay, 7: BAMG,
                          8: Frontal-Delaunay for Quads, 9: Packing of Parallelograms)
        :param interactive: Enable an interactive session using GMSH GUI.

        :type shp: ada.Shape
        :rtype: ada.FEM
        """
        import gmsh

        gmsh.option.setNumber("Mesh.Algorithm", mesh_algo)
        gmsh.option.setNumber("Mesh.ElementOrder", order)
        gmsh.option.setNumber("Mesh.SecondOrderIncomplete", 1)
        gmsh.option.setNumber("Mesh.Smoothing", 3)
        gmsh.option.setNumber("Geometry.Tolerance", 1e-3)

        gmsh.model.add(shp.name)
        shp.to_stp("temp_occ_in", os.path.join(self.work_dir))
        gmsh.open("gmsh/temp_occ_in.stp")

        model = gmsh.model
        factory = model.geo

        factory.synchronize()

        model.mesh.setRecombine(2, 1)
        model.mesh.generate(max_dim)
        model.mesh.removeDuplicateNodes()

        if interactive:
            gmsh.fltk.run()

    def _create_plate_geom(self, pl, beams, size=0.1, interactive=False):
        """

        :param pl:
        :param beams:
        :param size:
        :type pl: ada.Plate
        """
        import gmsh

        corners = [n for n in pl.poly.points3d]
        corners += [corners[0]]
        if clockwise:
            corners.reverse()

        def is_on_line(data):
            """
            Evaluate intersection point between two lines

            :param data:
            :return:
            """
            l, bm = data
            A, B = np.array(l[0]), np.array(l[1])
            AB = B - A
            C = bm.n1.p
            D = bm.n2.p
            CD = D - C

            if (vector_length(A - C) < 1e-5) is True and (vector_length(B - D) < 1e-5) is True:
                return None

            s, t = intersect_calc(A, C, AB, CD)
            AB_ = A + s * AB
            CD_ = C + t * CD
            if (vector_length(AB_ - CD_) < 1e-4) is True and s not in (0.0, 1.0):
                return list(AB_), bm
            else:
                return None

        # Evalute Corner Points
        all_points = []
        crossing_beams = []
        for s, e in zip(corners[:-1], corners[1:]):
            li = (s, e)
            all_points.append(s)
            res = [x for x in map(is_on_line, [(li, bm) for bm in beams]) if x is not None]
            add_points = [r[0] for r in res]
            crossing_beams += filter(lambda x: x not in crossing_beams, [r[1] for r in res])
            sorted_li = list(sorted(add_points, key=lambda x: vector_length(np.array(x) - np.array(s))))
            all_points += sorted_li
            all_points.append(e)

        cp = []
        for i in all_points:
            cp += [gmsh.model.geo.addPoint(*i, size)]

        lines = []
        for c1, c2 in zip(cp[:-1], cp[1:]):
            lines.append(gmsh.model.geo.addLine(c1, c2))

        gmsh.model.geo.removeAllDuplicates()
        d = gmsh.model.geo.addCurveLoop(lines)
        surf = gmsh.model.geo.addPlaneSurface([d])

        gmsh.model.geo.synchronize()

        intersec_geom = []
        for bm in crossing_beams:
            p1, p2 = bm.n1.p, bm.n2.p
            s = self.get_point(p1)
            e = self.get_point(p2)

            if len(e) == 0 or len(s) == 0:
                continue
                # raise ValueError("Point not found")
            intersec_geom += [gmsh.model.geo.addLine(s[0][1], e[0][1])]

        gmsh.model.geo.synchronize()
        gmsh.model.mesh.embed(1, intersec_geom, 2, surf)

        if interactive:
            gmsh.fltk.run()

        self._pl_map[surf] = pl
        self._bm_map.update({i: j for i, j in zip(intersec_geom, crossing_beams)})

    def _create_bm_geom(self, bm, size):
        """

        :param bm:
        :type bm: ada.Beam
        """
        import gmsh

        p1, p2 = bm.n1.p, bm.n2.p
        s = self.get_point(p1)
        e = self.get_point(p2)
        if len(s) == 0:
            s = [(0, gmsh.model.geo.addPoint(*p1.tolist(), size))]
        if len(e) == 0:
            e = [(0, gmsh.model.geo.addPoint(*p2.tolist(), size))]

        line = gmsh.model.geo.addLine(s[0][1], e[0][1])
        if line in self._bm_map.keys():
            raise ValueError("This should not happen!")
        self._bm_map[line] = bm

    def get_beam_elements(self, li, order):
        """

        :param li:
        :param order:
        :return:
        """
        import gmsh

        model = gmsh.model
        bm = self._bm_map[li]
        segments = model.mesh.getElements(1, li)[1][0]
        fem_nodes = model.mesh.getElements(1, li)[2][0]
        elem_types, elem_tags, elem_node_tags = gmsh.model.mesh.getElements(1, li)
        face, dim, morder, numv, parv, _ = gmsh.model.mesh.getElementProperties(elem_types[0])

        def make_elem(j):
            no = []
            for i in range(numv):
                p1 = fem_nodes[numv * j + i]
                p1_co = gmsh.model.mesh.getNode(p1)[0]
                no.append(Node(p1_co, p1))
            if len(no) == 3:
                myorder = [0, 2, 1]
                no = [no[i] for i in myorder]
            bm_el_type = "B31" if order == 1 else "B32"

            return Elem(segments[j], no, bm_el_type, parent=self._part.fem)

        elements = list(map(make_elem, range(len(segments))))

        femset = FemSet(f"el{bm.name}_set", elements, "elset", parent=self._part.fem)
        self._part.fem.sets.add(femset)
        self._part.fem.add_section(
            FemSection(
                f"d{bm.name}_sec",
                "beam",
                femset,
                bm.material,
                bm.section,
                bm.ori[2],
                metadata=dict(beam=bm, numel=len(elements)),
            )
        )
        return elements

    def get_shell_elements(self, sh, order):
        """

        :param sh:
        :param order:
        :return:
        """
        import gmsh

        pl = self._pl_map[sh]
        assert isinstance(pl, Plate)
        try:
            get_elems = gmsh.model.mesh.getElements(2, sh)
            segments = get_elems[1][0]
        except BaseException as e:
            logging.debug(e)
            return []
        fem_nodes = gmsh.model.mesh.getElements(2, sh)[2][0]
        elemTypes, elemTags, elemNodeTags = gmsh.model.mesh.getElements(2, sh)
        face, dim, morder, numv, parv, _ = gmsh.model.mesh.getElementProperties(elemTypes[0])

        elem_type = gmsh_map[face]

        def make_elem(j):
            el_id = segments[j]
            nonlocal elem_type
            nonlocal numv
            nonlocal fem_nodes
            no = []
            for i in range(numv):
                p1 = fem_nodes[numv * j + i]
                p1_co = gmsh.model.mesh.getNode(p1)[0]
                no.append(Node(p1_co, p1))

            return Elem(el_id, no, elem_type, parent=self._part.fem)

        elements = list(map(make_elem, range(len(segments))))

        femset = FemSet(f"el{pl.name}_set", elements, "elset")
        self._part.fem.add_set(femset)
        self._part.fem.add_section(
            FemSection(
                f"sh{pl.name}_sec",
                "shell",
                femset,
                pl.material,
                local_z=pl.n,
                thickness=pl.t,
                int_points=5,
                metadata=dict(beam=pl, numel=len(elements)),
            )
        )
        return elements

    def get_point(self, p):
        import gmsh

        tol = self._tol
        tol_vec = np.array([tol, tol, tol])
        lower = np.array(p) - tol_vec
        upper = np.array(p) + tol_vec
        return gmsh.model.getEntitiesInBoundingBox(*lower.tolist(), *upper.tolist(), 0)

    def get_mesh_nodes(self, e):
        """

        :param e:
        :return:
        """
        import gmsh

        nco = gmsh.model.mesh.getNode(e)
        return Node([roundoff(x) for x in nco[0]], e, parent=self._part.fem)

    @property
    def work_dir(self):
        return self._work_dir


def get_point(gmsh, p, tol=1e-5):
    """

    :param gmsh:
    :type gmsh: gmsh
    :param p:
    :param tol:
    :return: Entity Bounding box
    """
    tol_vec = np.array([tol, tol, tol])
    lower = np.array(p) - tol_vec
    upper = np.array(p) + tol_vec
    return gmsh.model.getEntitiesInBoundingBox(*lower.tolist(), *upper.tolist(), 0)


def get_nodes_and_elements(gmsh, fem=None, fem_set_name="all_elements"):
    """

    :param gmsh:
    :type gmsh: gmsh
    :param fem:
    :type fem: ada.fem.FEM
    :param fem_set_name:
    :type fem_set_name: str
    """
    from ada.fem import FEM

    fem = FEM("AdaFEM") if fem is None else fem

    nodes = list(gmsh.model.mesh.getNodes(-1, -1))
    # Get nodes
    fem._nodes = Nodes(
        [
            Node(
                [roundoff(x) for x in gmsh.model.mesh.getNode(n)[0]],
                n,
                parent=fem,
            )
            for n in nodes[0]
        ],
        parent=fem,
    )

    # Get elements
    elemTypes, elemTags, elemNodeTags = gmsh.model.mesh.getElements(2, -1)
    elements = []
    for k, element_list in enumerate(elemTags):
        face, dim, morder, numv, parv, _ = gmsh.model.mesh.getElementProperties(elemTypes[k])
        elem_type = gmsh_map[face]
        for j, eltag in enumerate(element_list):
            nodes = []
            for i in range(numv):
                idtag = numv * j + i
                p1 = elemNodeTags[k][idtag]
                nodes.append(fem.nodes.from_id(p1))

            el = Elem(eltag, nodes, elem_type, parent=fem)
            elements.append(el)
    fem._elements = FemElements(elements, fem_obj=fem)
    femset = FemSet(fem_set_name, elements, "elset")
    fem.sets.add(femset)


def eval_thick_normal_from_cog_of_beam_plate(beam, cog):
    """

    :param beam:
    :param cog:
    :type beam: ada.Beam
    :return:
    """
    from ada.core.utils import vector_length
    from ada.sections import SectionCat

    if SectionCat.is_circular_profile(beam.section.type):
        tol = beam.section.r / 8
    else:
        tol = beam.section.h / 8
    t, n, c = None, None, None
    xdir, ydir, zdir = beam.ori

    n1 = beam.n1.p
    n2 = beam.n2.p
    h = beam.section.h
    w_btn = beam.section.w_btn
    w_top = beam.section.w_top

    if beam.section.type in SectionCat.iprofiles + SectionCat.igirders:
        p11 = n1 + zdir * h / 2
        p12 = p11 + ydir * w_top / 2
        p21 = n2 + zdir * h / 2
        p22 = p21 + ydir * w_top / 2

        p11btn = n1 - zdir * h / 2 + ydir * w_btn / 2
        p12btn = p11btn - ydir * w_top / 2
        p21btn = n2 - zdir * h / 2
        p22btn = p21btn + ydir * w_btn / 2

        web = (n1 + n2) / 2

        fl_top_right = (p11 + p12 + p21 + p22) / 4
        fl_top_left = fl_top_right - ydir * w_top / 2
        fl_btn_right = (p11btn + p12btn + p21btn + p22btn) / 4
        fl_btn_left = fl_btn_right - ydir * w_btn / 2

        if vector_length(web - cog) < tol:
            t, n, c = beam.section.t_w, ydir, "web"

        for x in [fl_top_right, fl_top_left]:
            if vector_length(x - cog) < tol:
                t, n, c = beam.section.t_ftop, zdir, "top_fl"

        for x in [fl_btn_right, fl_btn_left]:
            if vector_length(x - cog) < tol:
                t, n, c = beam.section.t_fbtn, zdir, "btn_fl"

        if t is None:
            raise ValueError("The thickness is not valid")

    return t, n, c


def _init_gmsh_session():
    import gmsh

    gmsh_session = gmsh
    try:
        gmsh_session.finalize()
    except BaseException as e:
        logging.debug(e)
    gmsh_session.initialize()
    gmsh_session.option.setNumber("General.Terminal", 1)
    return gmsh_session


class GmshSession:
    def __init__(self, persist=True, geom_repr="shall", settings=None):
        print("init method called")
        self.gmsh = None
        self.settings = settings
        self.geom_repr = geom_repr
        self.persist = persist

    def run(self, function, *args, **kwargs):
        print("run function")
        res = function(self.gmsh, *args, **kwargs)
        if self.persist is False:
            self.gmsh.finalize()
            self.gmsh.initialize()
            self._add_settings()
        return res

    def _add_settings(self):
        if self.settings is not None:
            for setting, value in self.settings.items():
                self.gmsh.option.setNumber(setting, value)

    def __enter__(self):
        print("Starting GMSH session")
        import gmsh

        self.gmsh = gmsh
        self.gmsh.initialize()
        self._add_settings()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        print("Closing GMSH")
        self.gmsh.finalize()
